"""
All of the base-level stuff for plotting.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2007-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from matplotlib import figure
import shutil
import tempfile
import numpy as np
import os

from yt.funcs import *

from yt.config import ytcfg
from yt.data_objects.profiles import \
    BinnedProfile1D, \
    BinnedProfile2D
from yt.utilities.definitions import \
    axis_names, inv_axis_names, x_dict, y_dict
from .plot_types import \
    FixedResolutionPlot, \
    PCSlicePlot, \
    PCSlicePlotNaturalNeighbor, \
    PCProjectionPlot, \
    PCProjectionPlotNaturalNeighbor, \
    CuttingPlanePlot, \
    ParticlePlot, \
    ProfilePlot, \
    Profile1DPlot, \
    PhasePlot, \
    LineQueryPlot, \
    ScatterPlot
from yt.utilities.minimal_representation import \
    MinimalImageCollectionData

# No better place to put this
def concatenate_pdfs(output_fn, input_fns):
    from pyPdf import PdfFileWriter, PdfFileReader
    outfile = PdfFileWriter()
    for fn in input_fns:
        infile = PdfFileReader(open(fn, 'rb'))
        outfile.addPage(infile.getPage(0))
    outfile.write(open(output_fn, "wb"))

class ImageCollection(object):
    def __init__(self, pf, name):
        self.pf = pf
        self.name = name
        self.images = []
        self.image_metadata = []

    def add_image(self, fn, descr):
        self.image_metadata.append(descr)
        self.images.append((os.path.basename(fn), np.fromfile(fn, dtype='c')))

class PlotCollection(object):
    r"""The PlotCollection object was created to ease the creation of multiple
    slices, projections and so forth made from a single parameter file.
    The concept is that when the width on one image changes, it should
    change on all the others.  The PlotCollection can create all plot types
    available in yt.

    Parameters
    ----------
    pf : `StaticOutput`
        The parameter file from which all the plots will be created.
    center : array_like, optional
        The 'center' supplied to plots like sphere plots, slices, and so
        on.  Should be 3 elements.  Defaults to the point of maximum
        density.
    Long_variable_name : {'hi', 'ho'}, optional
        Choices in brackets, default first when optional.

    Notes
    -----
    This class is the primary entry point to creating plots, but it is not
    the only entry point.  Additionally, creating a PlotCollection should
    be a "cheap" operation.

    You may iterate over the plots in the PlotCollection, via something
    like:

    >>> pc = PlotCollection(pf)
    >>> for p in pc: print p

    Examples
    --------

    >>> pc = PlotCollection(pf, center=[0.5, 0.5, 0.5])
    >>> pc.add_slice("Density", 'x')
    >>> pc.save()

    """

    __id_counter = 0
    def __init__(self, pf, center=None):
        self.plots = []
        self.pf = pf
        if center == None:
            v,self.c = pf.h.find_max("Density") # @todo: ensure no caching
        elif center == "center" or center == "c":
            self.c = (pf.domain_right_edge + pf.domain_left_edge)/2.0
        else:
            self.c = np.array(center, dtype='float64')
        mylog.info("Created plot collection with default plot-center = %s",
                    list(self.c))

    def __iter__(self):
        for p in self.plots:
            yield p

    @property
    def _mrep(self):
        ic = ImageCollection(self.pf, "Plot Collection with center %s" % self.c)
        dd = tempfile.mkdtemp()
        fns = self.save(os.path.join(dd, "temp"))
        for fn, p in zip(fns, self.plots):
            ic.add_image(fn, p._pretty_name())
        shutil.rmtree(dd)
        return MinimalImageCollectionData(ic)

    def hub_upload(self):
        self._mrep.upload()

    def save(self, basename=None, format="png", override=False, force_save=False):
        r"""Save out all the plots hanging off this plot collection, using
        generated names.

        This function will create names for every plot that belongs to the
        PlotCollection and save them out.  The names will, by default, be
        prefixed with the name of the affiliate parameter file, and the file
        names should indicate clearly what each plot represents.

        Parameters
        ----------
        basename : string, optional
            The prefix for all of the plot filenames.
        format : string, optional
            The plot file format.  Can be 'png', 'pdf', 'eps', 'jpg', or
            anything else that matplotlib understands.
        override : boolean
            If this is true, then no generated filenames will be appended to
            the base name.  You probably don't want this.
        force_save : boolean
            In parallel, only the root task (proc 0) saves an image, unless
            this is set to True.

        Returns
        -------
        items : string
            This function returns a list of the filenames created.

        Examples
        --------

        >>> fns = pc.save()
        >>> for fn in fns: print "Saved", fn
        """
        if basename is None: basename = str(self.pf)
        fn = []
        for plot in self.plots:
            fn.append(plot.save_image(basename, format=format, 
                      override=override, force_save=force_save))
            mylog.info("Saved %s", fn[-1])
        if ytcfg.getboolean("yt", "__withinreason"):
            from yt.gui.reason.bottle_mods import PayloadHandler
            import base64
            ph = PayloadHandler()
            for f in fn:
                if not f.endswith('png'): continue
                img_data = base64.b64encode(open(f,'rb').read())
                payload = {'type':'png_string',
                           'image_data':img_data}
                ph.add_payload(payload)
        return fn

    def set_xlim(self, xmin, xmax):
        r"""Set the x-limits of all plots.

        set_xlim on all plots is called with the parameters passed to this
        function.

        Parameters
        ----------
        xmin : float
            The left boundary for the x axis.
        xmax : float
            The right boundary for the x axis.
        """
        for plot in self.plots:
            plot.set_xlim(xmin, xmax)

    def set_ylim(self, ymin, ymax):
        r"""Set the y-limits of all plots.

        set_ylim on all plots is called with the parameters passed to this
        function.

        Parameters
        ----------
        ymin : float
            The left boundary for the x axis.
        ymax : float
            The right boundary for the x axis.
        """
        for plot in self.plots:
            plot.set_ylim(ymin, ymax)

    def set_zlim(self, zmin, zmax, *args, **kwargs):
        """
        Set the limits of the colorbar. 'min' or 'max' are possible inputs 
        when combined with dex=value, where value gives the maximum number of 
        dex to go above/below the min/max.  If value is larger than the true
        range of values, min/max are limited to true range.

        Only ONE of the following options can be specified. If all 3 are
        specified, they will be used in the following precedence order:

        * ``ticks`` - a list of floating point numbers at which to put ticks
        * ``minmaxtick`` - display DEFAULT ticks with min & max also displayed
        * ``nticks`` - if ticks not specified, can automatically determine a
          number of ticks to be evenly spaced in log space
        """
        for plot in self.plots:
            plot.set_autoscale(False)
            plot.set_zlim(zmin, zmax, *args, **kwargs)

    def set_lim(self, lim):
        r"""Set the x- and y-limits of all plots.

        set_xlim on all plots is called with the parameters passed to this
        function, and then set_ylim is called.

        Parameters
        ----------
        lim : tuple of floats
            (xmin, xmax, ymin, ymax)
        """
        for plot in self.plots:
            plot.set_xlim(*lim[:2])
            plot.set_ylim(*lim[2:])

    def autoscale(self):
        r"""Turn on autoscaling on all plots.

        This has the same effect as:

        >>> for p in pc: p.set_autoscale(True)

        By default, all plots are autoscaled until the colorbar is set
        manually.  This turns autoscaling back on.  The colors may not be
        updated unless _redraw_image is called on the plots, which should occur
        with a change in the width or saving of images.
        """
        for plot in self.plots:
            plot.set_autoscale(True)

    def set_width(self, width, unit):
        r"""Change the width of all image plots.

        This function changes all the widths of the image plots (but notably
        not any phase plots or profile plots) to be a given physical extent.

        Parameters
        ----------
        width : float
            The numeric value of the new width.
        unit : string
            The unit in which the given width is expressed.
        """
        for plot in self.plots:
            plot.set_width(width, unit)

    def set_cmap(self, cmap):
        r"""Change the colormap of all plots.

        This function will update the colormap on all plots for which a
        colormap makes sense.  The colors may not be updated unless
        _redraw_image is called on the plots, which should occur with a change
        in the width or saving of images.

        Parameters
        ----------
        cmap : string
            An acceptable colormap.  See either yt.visualization.color_maps or
            http://www.scipy.org/Cookbook/Matplotlib/Show_colormaps .
        """
        for plot in self.plots:
            plot.set_cmap(cmap)

    def switch_field(self, field):
        r"""Change the displayed of all image plots.

        All images that display a field -- slices, cutting planes, projections
        -- will be switched to display the specified field.  For projections,
        this will re-generate the projection, if it is unable to load the
        projected field off-disk.

        Parameters
        ----------
        field : string
            Any field that can be generated or read from disk.
        """
        for plot in self.plots:
            plot.switch_z(field)
    switch_z = switch_field

    def _add_plot(self, plot):
        r"""This function adds a plot to the plot collection.

        This function is typically used internally to add a plot on to the
        current list of plots.  However, if you choose to manually create a
        plot, this can be used to add it to a collection for convenient
        modification.

        Parameters
        ----------
        plot : `yt.visualization.plot_types.RavenPlot`
            A plot, which will be appended to the list of plots handled by this
            plot collection.

        Returns
        -------
        plot : `yt.visualization.plot_types.RavenPlot`
            The plot handed to the function is passed back through.  This is
            unnecessary, but is done for historical reasons.
        """
        self.plots.append(plot)
        return plot

    def add_slice(self, field, axis, coord=None, center=None,
                 use_colorbar=True, figure = None, axes = None, fig_size=None,
                 periodic = True, obj = None, field_parameters = None):
        r"""Create a slice, from that a slice plot, and add it to the current
        collection.

        This function will generate a `yt.data_objects.api.AMRSliceBase` from the given
        parameters.  This slice then gets passed to a `yt.visualization.plot_types.PCSlicePlot`, and
        the resultant plot is added to the current collection.  Various
        parameters allow control of the way the slice is displayed, as well as
        how the slice is generated.

        Parameters
        ----------
        field : string
            The initial field to slice and display.
        axis : int
            The axis along which to slice.  Can be 0, 1, or 2 or x, y, z.
        coord : float, optional
            The coordinate to place the slice at, along the slicing axis.
        center : array_like, optional
            The center to be used for things like radius and radial velocity.
            Defaults to the center of the plot collection.
        use_colorbar : bool, optional
            Whether we should leave room for and create a colorbar.
        figure : `matplotlib.figure.Figure`, optional
            The figure onto which the axes will be placed.  Typically not used
            unless *axes* is also specified.
        axes : `matplotlib.axes.Axes`, optional
            The axes object which will be used to create the image plot.
            Typically used for things like multiplots and the like.
        fig_size : tuple of floats
            This parameter can act as a proxy for the manual creation of a
            figure.  By specifying it, you can create plots with an arbitrarily
            large or small size.  It is in inches, defaulting to 100 dpi.
        periodic : boolean, optional
            By default, the slices are assumed to be periodic, and they will
            wrap around the edges.
        obj : `yt.data_objects.api.AMRSliceBase`, optional
            If you would like to use an existing slice, you may specify it
            here, in which case a new slice will not be created.
        field_parameters : dict, optional
            This set of parameters will be passed to the slice upon creation,
            which can be used for passing variables to derived fields.

        Returns
        -------
        plot : `yt.visualization.plot_types.PCSlicePlot`
            The plot that has been added to the PlotCollection.

        See Also
        --------
        yt.data_objects.api.AMRSliceBase : This is the type created by this function and 
                                passed to the plot created here.

        Notes
        -----
        This is the primary mechanism for creating slice plots, and generating
        slice plots along multiple axes was the original purpose of the
        PlotCollection.

        Note that all plots can be modified.  See `callback_list` for more
        information.

        Examples
        --------

        >>> pf = load("RD0005-mine/RedshiftOutput0005")
        >>> pc = PlotCollection(pf, [0.5, 0.5, 0.5])
        >>> p = pc.add_slice("Density", 'x')
        """
        axis = fix_axis(axis)
        if center == None:
            center = self.c
        if coord == None:
            coord = center[axis]
        if obj is None:
            if field_parameters is None: field_parameters = {}
            obj = self.pf.hierarchy.slice(axis, coord, field,
                            center=center, **field_parameters)
        p = self._add_plot(PCSlicePlot(
                         obj, field, use_colorbar=use_colorbar,
                         axes=axes, figure=figure,
                         size=fig_size, periodic=periodic))
        mylog.info("Added slice of %s at %s = %s with 'center' = %s", field,
                    axis_names[axis], coord, list(center))
        p["Axis"] = axis_names[axis]
        return p

    def add_particles(self, axis, width, p_size=1.0, col='k', stride=1.0,
                      data_source=None, figure=None, axes=None):
        r"""Create a plot of a thick slab of particles.

        This function will generate a `yt.data_objects.api.AMRRegionBase` from the given
        parameters, and all particles which are within that region will be
        plotted.

        Parameters
        ----------
        axis : int
            The axis along which to create the thick slab.  Can be 0, 1, or 2
            or x, y, z.
        width : float
            The width of the thick slab, in code units, from which particles
            will be plotted.
        p_size : float, optional
            The size of the points to be used to represent the particles, in
            pixels.
        col : color, optional
            Specified in matplotlib color specifications, the color that
            particles should be.
        stride : float, optional
            The stride through the particles to plot.  Used to plot every
            fifth, every tenth, etc.  Note that the sorted order of particles
            may result in a biased selection of particles.
        data_source : `yt.data_objects.api.AMRData`, optional
            If specified, this will be the data source used for obtaining
            particles.
        figure : `matplotlib.figure.Figure`, optional
            The figure onto which the axes will be placed.  Typically not used
            unless *axes* is also specified.
        axes : `matplotlib.axes.Axes`, optional
            The axes object which will be used to create the image plot.
            Typically used for things like multiplots and the like.

        Returns
        -------
        plot : `yt.visualization.plot_types.ParticlePlot`
            The plot that has been added to the PlotCollection.

        Notes
        -----
        This plot type can be very expensive, and does not necessarily produce
        the best visual results.  Plotting a large number of particles can be
        very tricky, and often it's much better to instead use a slice or a
        (thin) projection of deposited density, like particle_density_pyx.

        Examples
        --------

        >>> pf = load("RD0005-mine/RedshiftOutput0005")
        >>> pc = PlotCollection(pf, [0.5, 0.5, 0.5])
        >>> p = pc.add_particles(0, 1.0)
        """
        axis = fix_axis(axis)
        LE = self.pf.domain_left_edge.copy()
        RE = self.pf.domain_right_edge.copy()
        LE[axis] = self.c[axis] - width/2.0
        RE[axis] = self.c[axis] + width/2.0
        if data_source is None: data_source = self.pf.h.region(self.c, LE, RE)
        data_source.axis = axis
        p = self._add_plot(ParticlePlot(data_source, axis,
                                        width, p_size, col, stride, figure,
                                        axes))
        p["Axis"] = axis_names[axis]
        return p

    def add_cutting_plane(self, field, normal,
                          center=None, use_colorbar=True,
                          figure = None, axes = None, fig_size=None, obj=None,
                           field_parameters = None):
        r"""Create a cutting plane, from that a plot, and add it to the current
        collection.

        A cutting plane is an oblique slice through the simulation volume,
        oriented by a specified normal vector that is perpendicular to the
        image plane.  This function will generate a
        `yt.data_objects.api.AMRCuttingPlaneBase` from the given parameters.  This cutting
        plane then gets passed to a `yt.visualization.plot_types.CuttingPlanePlot`, and the
        resultant plot is added to the current collection.  Various parameters
        allow control of the way the slice is displayed, as well as how the
        plane is generated.

        Parameters
        ----------
        field : string
            The initial field to slice and display.
        normal : array_like
            The vector that defines the desired plane.  For instance, the
            angular momentum of a sphere.
        center : array_like, optional
            The center to be used for things like radius and radial velocity.
            Defaults to the center of the plot collection.
        use_colorbar : bool, optional
            Whether we should leave room for and create a colorbar.
        figure : `matplotlib.figure.Figure`, optional
            The figure onto which the axes will be placed.  Typically not used
            unless *axes* is also specified.
        axes : `matplotlib.axes.Axes`, optional
            The axes object which will be used to create the image plot.
            Typically used for things like multiplots and the like.
        fig_size : tuple of floats
            This parameter can act as a proxy for the manual creation of a
            figure.  By specifying it, you can create plots with an arbitrarily
            large or small size.  It is in inches, defaulting to 100 dpi.
        obj : `AMRCuttingPlaneBase`, optional
            If you would like to use an existing cutting plane, you may specify
            it here, in which case a new cutting plane will not be created.
        field_parameters : dict, optional
            This set of parameters will be passed to the cutting plane upon
            creation, which can be used for passing variables to derived
            fields.

        Returns
        -------
        plot : `yt.visualization.plot_types.CuttingPlanePlot`
            The plot that has been added to the PlotCollection.

        See Also
        --------
        yt.data_objects.api.AMRCuttingPlaneBase : This is the type created by this function.

        Notes
        -----
        This is the primary mechanism for creating cutting plane plots.  Note
        that they are somewhat slow, but useful to orient the image in an
        arbitrary direction.

        Note that all plots can be modified.  See `callback_list` for more
        information.

        Examples
        --------

        Here's a simple mechanism for getting the angular momentum of a
        collapsing cloud and generating a cutting plane aligned with the
        angular momentum vector.

        >>> pf = load("RD0005-mine/RedshiftOutput0005")
        >>> v, c = pf.h.find_max("Density")
        >>> sp = pf.h.sphere(c, 1000.0/pf['au'])
        >>> L = sp.quantities["AngularMomentumVector"]()
        >>> pc = PlotCollection(pf)
        >>> p = pc.add_cutting_plane("Density", L)
        """
        if center == None:
            center = self.c
        if not obj:
            if field_parameters is None: field_parameters = {}
            cp = self.pf.hierarchy.cutting(normal, center, field,
                    **field_parameters)
        else:
            cp = obj
        p = self._add_plot(CuttingPlanePlot(cp, field,
                         use_colorbar=use_colorbar, axes=axes, figure=figure,
                         size=fig_size))
        mylog.info("Added plane of %s with 'center' = %s and normal = %s", field,
                    list(center), list(normal))
        p["Axis"] = "CuttingPlane"
        return p

    def add_fixed_res_cutting_plane(self, field, normal, width, res=512,
             center=None, use_colorbar=True, figure = None, axes = None,
             fig_size=None, obj=None, field_parameters = None):
        r"""Create a fixed resolution cutting plane, from that a plot, and add
        it to the current collection.

        A cutting plane is an oblique slice through the simulation volume,
        oriented by a specified normal vector that is perpendicular to the
        image plane.  This function will slice through, but instead of
        retaining all the data necessary to rescale the cutting plane at any
        width, it only retains the pixels for a single width.  This function
        will generate a `yt.data_objects.api.AMRFixedResCuttingPlaneBase` from the given
        parameters.  This image buffer then gets passed to a
        `yt.visualization.plot_types.FixedResolutionPlot`, and the resultant plot is added to the
        current collection.  Various parameters allow control of the way the
        slice is displayed, as well as how the plane is generated.

        Parameters
        ----------
        field : string
            The initial field to slice and display.
        normal : array_like
            The vector that defines the desired plane.  For instance, the
            angular momentum of a sphere.
        width : float
            The width, in code units, of the image plane.
        res : int
            The returned image buffer must be square; this number is how many
            pixels on a side it will have.
        center : array_like, optional
            The center to be used for things like radius and radial velocity.
            Defaults to the center of the plot collection.
        use_colorbar : bool, optional
            Whether we should leave room for and create a colorbar.
        figure : `matplotlib.figure.Figure`, optional
            The figure onto which the axes will be placed.  Typically not used
            unless *axes* is also specified.
        axes : `matplotlib.axes.Axes`, optional
            The axes object which will be used to create the image plot.
            Typically used for things like multiplots and the like.
        fig_size : tuple of floats
            This parameter can act as a proxy for the manual creation of a
            figure.  By specifying it, you can create plots with an arbitrarily
            large or small size.  It is in inches, defaulting to 100 dpi.
        obj : `AMRCuttingPlaneBase`, optional
            If you would like to use an existing cutting plane, you may specify
            it here, in which case a new cutting plane will not be created.
        field_parameters : dict, optional
            This set of parameters will be passed to the cutting plane upon
            creation, which can be used for passing variables to derived
            fields.

        Returns
        -------
        plot : `yt.visualization.plot_types.FixedResolutionPlot`
            The plot that has been added to the PlotCollection.

        See Also
        --------
        yt.data_objects.api.AMRFixedResCuttingPlaneBase : This is the type created by this
                                               function.

        Examples
        --------

        Here's a simple mechanism for getting the angular momentum of a
        collapsing cloud and generating a cutting plane aligned with the
        angular momentum vector.

        >>> pf = load("RD0005-mine/RedshiftOutput0005")
        >>> v, c = pf.h.find_max("Density")
        >>> sp = pf.h.sphere(c, 1000.0/pf['au'])
        >>> L = sp.quantities["AngularMomentumVector"]()
        >>> pc = PlotCollection(pf)
        >>> p = pc.add_fixed_res_cutting_plane("Density", L, 1000.0/pf['au'])
        """
        if center == None:
            center = self.c
        if not obj:
            if field_parameters is None: field_parameters = {}
            data = self.pf.hierarchy.fixed_res_cutting \
                 (normal, center, width, res, **field_parameters)
            #data = frc[field]
        else:
            data = obj
        p = self._add_plot(FixedResolutionPlot(data, field,
                         use_colorbar=use_colorbar, axes=axes, figure=figure,
                         size=fig_size))
        mylog.info("Added fixed-res plane of %s with 'center' = %s and "
                   "normal = %s", field, list(center), list(normal))
        p["Axis"] = "CuttingPlane"
        return p

    def add_projection(self, field, axis,  weight_field=None,
                       data_source = None,
                       center=None, use_colorbar=True,
                       figure = None, axes = None, fig_size=None,
                       periodic = True, obj = None, field_parameters = None):
        r"""Create a projection, from that a projection plot, and add it to the
        current collection.

        This function will generate a `yt.data_objects.api.AMRProjBase` from the given
        parameters.  This projection then gets passed to a
        `yt.visualization.plot_types.PCProjectionPlot`, and the resultant plot is added to the
        current collection.  Various parameters allow control of the way the
        slice is displayed, as well as how the slice is generated.

        Parameters
        ----------
        field : string
            The initial field to slice and display.
        axis : int
            The axis along which to slice.  Can be 0, 1, or 2 or x, y, z.
        data_source : `yt.data_objects.api.AMRData`
            This is a data source respecting the `AMRData` protocol (i.e., it
            has grids and so forth) that will be used as input to the
            projection.
        weight_field : string
            If specified, this will be the weighting field and the resultant
            projection will be a line-of-sight average, defined as sum( f_i *
            w_i * dl ) / sum( w_i * dl )
        center : array_like, optional
            The center to be used for things like radius and radial velocity.
            Defaults to the center of the plot collection.
        use_colorbar : bool, optional
            Whether we should leave room for and create a colorbar.
        figure : `matplotlib.figure.Figure`, optional
            The figure onto which the axes will be placed.  Typically not used
            unless *axes* is also specified.
        axes : `matplotlib.axes.Axes`, optional
            The axes object which will be used to create the image plot.
            Typically used for things like multiplots and the like.
        fig_size : tuple of floats
            This parameter can act as a proxy for the manual creation of a
            figure.  By specifying it, you can create plots with an arbitrarily
            large or small size.  It is in inches, defaulting to 100 dpi.
        periodic : boolean, optional
            By default, the slices are assumed to be periodic, and they will
            wrap around the edges.
        obj : `yt.data_objects.api.AMRProjBase`, optional
            If you would like to use an existing projection, you may specify it
            here, in which case a new projection will not be created.  If this
            option is specified the options data_source, weight_field and
            field_parameters will be ignored.
        field_parameters : dict, optional
            This set of parameters will be passed to the slice upon creation,
            which can be used for passing variables to derived fields.

        Returns
        -------
        plot : `yt.visualization.plot_types.PCProjectionPlot`
            The plot that has been added to the PlotCollection.

        See Also
        --------
        yt.data_objects.api.AMRProjBase : This is the type created by this function and 
                               passed to the plot created here.

        Notes
        -----
        This is the primary mechanism for creating projection plots, and
        generating projection plots along multiple axes was the original
        purpose of the PlotCollection.

        Note that all plots can be modified.  See `callback_list` for more
        information.

        Examples
        --------

        >>> pf = load("RD0005-mine/RedshiftOutput0005")
        >>> pc = PlotCollection(pf, [0.5, 0.5, 0.5])
        >>> p = pc.add_projection("Density", 'x', "Density")
        """
        axis = fix_axis(axis)
        if field_parameters is None: field_parameters = {}
        if center == None:
            center = self.c
        if obj is None:
            obj = self.pf.hierarchy.proj(axis, field, weight_field,
                                         source = data_source, center=center,
                                         **field_parameters)
        p = self._add_plot(PCProjectionPlot(obj, field,
                         use_colorbar=use_colorbar, axes=axes, figure=figure,
                         size=fig_size, periodic=periodic))
        p["Axis"] = axis_names[axis]
        return p

    def add_thin_projection(self, field, axis, thickness,
                       weight_field=None, center=None, use_colorbar=True,
                       figure = None, axes = None, fig_size=None,
                       periodic = True, field_parameters = None):
        r"""Create a projection through a thin slice of the domain, from that a
        projection plot, and add it to the current collection.

        This function will generate a rectangular prism region and supply it to
        a`yt.data_objects.api.AMRProjBase` from the given parameters.  This projection
        then gets passed to a `yt.visualization.plot_types.PCProjectionPlot`, and the resultant plot
        is added to the current collection.  Various parameters allow control
        of the way the slice is displayed, as well as how the slice is
        generated.  The center is used as the center of the thin projection.

        Parameters
        ----------
        field : string
            The initial field to slice and display.
        axis : int
            The axis along which to slice.  Can be 0, 1, or 2 or x, y, z.
        thickness : float
            In 'code units' this is the thickness of the region to be
            projected through.
        weight_field : string
            If specified, this will be the weighting field and the resultant
            projection will be a line-of-sight average, defined as sum( f_i *
            w_i * dl ) / sum( w_i * dl )
        center : array_like, optional
            The center to be used for things like radius and radial velocity.
            Defaults to the center of the plot collection.
        use_colorbar : bool, optional
            Whether we should leave room for and create a colorbar.
        figure : `matplotlib.figure.Figure`, optional
            The figure onto which the axes will be placed.  Typically not used
            unless *axes* is also specified.
        axes : `matplotlib.axes.Axes`, optional
            The axes object which will be used to create the image plot.
            Typically used for things like multiplots and the like.
        fig_size : tuple of floats
            This parameter can act as a proxy for the manual creation of a
            figure.  By specifying it, you can create plots with an arbitrarily
            large or small size.  It is in inches, defaulting to 100 dpi.
        periodic : boolean, optional
            By default, the slices are assumed to be periodic, and they will
            wrap around the edges.
        field_parameters : dict, optional
            This set of parameters will be passed to the slice upon creation,
            which can be used for passing variables to derived fields.

        Returns
        -------
        plot : `yt.visualization.plot_types.PCProjectionPlot`
            The plot that has been added to the PlotCollection.

        See Also
        --------
        yt.data_objects.api.AMRProjBase : This is the type created by this function and 
                               passed to the plot created here.

        Notes
        -----
        This is the primary mechanism for creating projection plots, and
        generating projection plots along multiple axes was the original
        purpose of the PlotCollection.

        Note that all plots can be modified.  See `callback_list` for more
        information.

        Examples
        --------

        >>> pf = load("RD0005-mine/RedshiftOutput0005")
        >>> pc = PlotCollection(pf, [0.5, 0.5, 0.5])
        >>> p = pc.add_thin_projection("Density", 0, 0.1, "Density")
        """
        axis = fix_axis(axis)
        if field_parameters is None: field_parameters = {}
        if center == None:
            center = self.c
        LE = self.pf.domain_left_edge.copy()
        RE = self.pf.domain_right_edge.copy()
        LE[axis] = RE[axis] = center[axis]
        LE[axis] -= thickness/2.0
        RE[axis] += thickness/2.0
        region = self.pf.h.region(center, LE, RE)
        obj = self.pf.hierarchy.proj(axis, field, weight_field,
                                     source = region, center=center,
                                     **field_parameters)
        p = self._add_plot(PCProjectionPlot(obj, field,
                         use_colorbar=use_colorbar, axes=axes, figure=figure,
                         size=fig_size, periodic=periodic))
        p["Axis"] = axis_names[axis]
        return p

    def add_profile_object(self, data_source, fields,
                           weight="CellMassMsun", accumulation=False,
                           x_bins=128, x_log=True, x_bounds=None,
                           lazy_reader=True, id=None,
                           figure=None, axes=None):
        r"""From an existing object, create a 1D, binned profile.

        This function will accept an existing `AMRData` source and from that,
        it will generate a `Binned1DProfile`, based on the specified options.
        This is useful if you have extracted a region, or if you wish to bin
        some set of massages data -- or even if you wish to bin anything other
        than a sphere.  The profile will be 1D, which means while it can have
        an arbitrary number of fields, those fields will all be binned based on
        a single field.

        Parameters
        ----------
        data_source : `yt.data_objects.api.AMRData`
            This is a data source respecting the `AMRData` protocol (i.e., it
            has grids and so forth) that will be used as input to the profile
            generation.
        fields : list of strings
            The first element of this list is the field by which we will bin;
            all subsequent fields will be binned and their profiles added to
            the underlying `BinnedProfile1D`.
        weight : string, default "CellMassMsun"
            The weighting field for an average.  This defaults to mass-weighted
            averaging.
        accumulation : boolean, optional
            If true, from the low-value to the high-value the values in all
            binned fields will be accumulated.  This is useful for instance
            when adding an unweighted CellMassMsun to a radial plot, as it will
            show mass interior to that radius.
        x_bins : int, optional
            How many bins should there be in the independent variable?
        x_log : boolean, optional
            Should the bin edges be log-spaced?
        x_bounds : tuple of floats, optional
            If specified, the boundary values for the binning.  If unspecified,
            the min/max from the data_source will be used.  (Non-zero min/max
            in case of log-spacing.)
        lazy_reader : boolean, optional
            If this is false, all of the data will be read into memory before
            any processing occurs.  It defaults to true, and grids are binned
            on a one-by-one basis.  Note that parallel computation requires
            this to be true.
        id : int, optional
            If specified, this will be the "semi-unique id" of the resultant
            plot.  This should not be set.
        figure : `matplotlib.figure.Figure`, optional
            The figure onto which the axes will be placed.  Typically not used
            unless *axes* is also specified.
        axes : `matplotlib.axes.Axes`, optional
            The axes object which will be used to create the image plot.
            Typically used for things like multiplots and the like.

        Returns
        -------
        plot : `yt.visualization.plot_types.ProfilePlot`
            The plot that has been added to the PlotCollection.

        See Also
        --------
        yt.data_objects.profiles.BinnedProfile1D : This is the object that does the
                                   transformation of raw data into a 1D
                                   profile.

        Examples
        --------

        >>> reg = pf.h.region([0.1, 0.2, 0.3], [0.0, 0.1, 0.2],
                              [0.2, 0.3, 0.4])
        >>> pc.add_profile_object(reg, ["Density", "Temperature"])
        """
        if x_bounds is None:
            x_min, x_max = data_source.quantities["Extrema"](
                            fields[0], non_zero = x_log,
                            lazy_reader=lazy_reader)[0]
        else:
            x_min, x_max = x_bounds
        profile = BinnedProfile1D(data_source,
                                  x_bins, fields[0], x_min, x_max, x_log,
                                  lazy_reader)
        if len(fields) > 1:
            profile.add_fields(fields[1:], weight=weight, accumulation=accumulation)
        if id is None: id = self._get_new_id()
        p = self._add_plot(Profile1DPlot(profile, fields, id,
                                                   axes=axes, figure=figure))
        return p

    def add_profile_sphere(self, radius, unit, fields, center = None,
                           weight="CellMassMsun", accumulation=False,
                           x_bins=128, x_log=True, x_bounds=None,
                           lazy_reader=True, id=None,
                           figure=None, axes=None):
        r"""From a description of a sphere, create a 1D, binned profile.

        This function will accept the radius of a sphere, and from that it will
        generate a `Binned1DProfile`, based on the specified options.  The
        profile will be 1D, which means while it can have an arbitrary number
        of fields, those fields will all be binned based on a single field.

        All subsequent parameters beyond "unit" will be passed verbatim to
        add_profile_object.

        Parameters
        ----------
        radius : float
            The radius of the sphere to generate.
        unit : string
            The unit in which the given radius is expressed.
        fields : list of strings
            The first element of this list is the field by which we will bin;
            all subsequent fields will be binned and their profiles added to
            the underlying `BinnedProfile1D`.
        center : array_like, optional
            The center to be used for things like radius and radial velocity.
            Defaults to the center of the plot collection.
        weight : string, default "CellMassMsun"
            The weighting field for an average.  This defaults to mass-weighted
            averaging.
        accumulation : boolean, optional
            If true, from the low-value to the high-value the values in all
            binned fields will be accumulated.  This is useful for instance
            when adding an unweighted CellMassMsun to a radial plot, as it will
            show mass interior to that radius.
        x_bins : int, optional
            How many bins should there be in the independent variable?
        x_log : boolean, optional
            Should the bin edges be log-spaced?
        x_bounds : tuple of floats, optional
            If specified, the boundary values for the binning.  If unspecified,
            the min/max from the data_source will be used.  (Non-zero min/max
            in case of log-spacing.)
        lazy_reader : boolean, optional
            If this is false, all of the data will be read into memory before
            any processing occurs.  It defaults to true, and grids are binned
            on a one-by-one basis.  Note that parallel computation requires
            this to be true.
        id : int, optional
            If specified, this will be the "semi-unique id" of the resultant
            plot.  This should not be set.
        figure : `matplotlib.figure.Figure`, optional
            The figure onto which the axes will be placed.  Typically not used
            unless *axes* is also specified.
        axes : `matplotlib.axes.Axes`, optional
            The axes object which will be used to create the image plot.
            Typically used for things like multiplots and the like.

        Returns
        -------
        plot : `yt.visualization.plot_types.ProfilePlot`
            The plot that has been added to the PlotCollection.  Note that the
            underlying sphere may be accessed as .data.data_source

        See Also
        --------
        yt.data_objects.profiles.BinnedProfile1D : This is the object that does the
                                   transformation of raw data into a 1D
                                   profile.
        yt.data_objects.api.AMRSphereBase : This is the object auto-generated by this
                                 function.

        Examples
        --------

        >>> pc.add_profile_sphere(1.0, 'kpc', ["Density", "Electron_Fraction"])
        """
        if center is None:
            center = self.c
        r = radius/self.pf[unit]
        sphere = self.pf.hierarchy.sphere(center, r)
        p = self.add_profile_object(sphere, fields, weight, accumulation,
                           x_bins, x_log, x_bounds, lazy_reader, id,
                           figure=figure, axes=axes)
        p["Width"] = radius
        p["Unit"] = unit
        p["Axis"] = None
        return p

    def add_phase_object(self, data_source, fields, cmap=None,
                               weight="CellMassMsun", accumulation=False,
                               x_bins=128, x_log=True, x_bounds=None,
                               y_bins=128, y_log=True, y_bounds=None,
                               lazy_reader=True, id=None,
                               axes = None, figure = None,
                               fractional=False):
        r"""From an existing object, create a 2D, binned profile.

        This function will accept an existing `AMRData` source and from that,
        it will generate a `Binned2DProfile`, based on the specified options.
        This is useful if you have extracted a region, or if you wish to bin
        some set of massages data -- or even if you wish to bin anything other
        than a sphere.  The profile will be 2D, which means while it can have
        an arbitrary number of fields, those fields will all be binned based on
        two fields.

        Parameters
        ----------
        data_source : `yt.data_objects.api.AMRData`
            This is a data source respecting the `AMRData` protocol (i.e., it
            has grids and so forth) that will be used as input to the profile
            generation.
        fields : list of strings
            The first element of this list is the field by which we will bin
            into the x-axis, the second is the field by which we will bin onto
            the y-axis.  All subsequent fields will be binned and their
            profiles added to the underlying `BinnedProfile2D`.
        cmap : string, optional
            An acceptable colormap.  See either yt.visualization.color_maps or
            http://www.scipy.org/Cookbook/Matplotlib/Show_colormaps .
        weight : string, default "CellMassMsun"
            The weighting field for an average.  This defaults to mass-weighted
            averaging.
        accumulation : list of booleans, optional
            If true, from the low-value to the high-value the values in all
            binned fields will be accumulated.  This is useful for instance
            when adding an unweighted CellMassMsun to a radial plot, as it will
            show mass interior to that radius.  The first value is for the
            x-axis, the second value for the y-axis.  Note that accumulation
            will only be along each row or column.
        x_bins : int, optional
            How many bins should there be in the x-axis variable?
        x_log : boolean, optional
            Should the bin edges be log-spaced?
        x_bounds : tuple of floats, optional
            If specified, the boundary values for the binning.  If unspecified,
            the min/max from the data_source will be used.  (Non-zero min/max
            in case of log-spacing.)
        y_bins : int, optional
            How many bins should there be in the y-axis variable?
        y_log : boolean, optional
            Should the bin edges be log-spaced?
        y_bounds : tuple of floats, optional
            If specified, the boundary values for the binning.  If unspecified,
            the min/max from the data_source will be used.  (Non-zero min/max
            in case of log-spacing.)
        lazy_reader : boolean, optional
            If this is false, all of the data will be read into memory before
            any processing occurs.  It defaults to true, and grids are binned
            on a one-by-one basis.  Note that parallel computation requires
            this to be true.
        id : int, optional
            If specified, this will be the "semi-unique id" of the resultant
            plot.  This should not be set.
        figure : `matplotlib.figure.Figure`, optional
            The figure onto which the axes will be placed.  Typically not used
            unless *axes* is also specified.
        axes : `matplotlib.axes.Axes`, optional
            The axes object which will be used to create the image plot.
            Typically used for things like multiplots and the like.
        fractional : boolean
            If true, the plot will be normalized to the sum of all the binned
            values.

        Returns
        -------
        plot : `yt.visualization.plot_types.PlotTypes.PhasePlot`
            The plot that has been added to the PlotCollection.

        See Also
        --------
        yt.data_objects.profiles.BinnedProfile2D : This is the object that does the
                                   transformation of raw data into a 1D
                                   profile.
        
        Examples
        --------
        This will show the mass-distribution in the Density-Temperature plane.
        
        >>> pf = load("RD0005-mine/RedshiftOutput0005")
        >>> reg = pf.h.region([0.1, 0.2, 0.3], [0.0, 0.1, 0.2],
        ...                   [0.2, 0.3, 0.4])
        >>> pc.add_phase_object(reg, ["Density", "Temperature", "CellMassMsun"],
        ...                     weight = None)
        """
        if x_bounds is None:
            x_min, x_max = data_source.quantities["Extrema"](
                                    fields[0], non_zero = x_log,
                                    lazy_reader=lazy_reader)[0]
        else:
            x_min, x_max = x_bounds
        if y_bounds is None:
            y_min, y_max = data_source.quantities["Extrema"](
                                    fields[1], non_zero = y_log,
                                    lazy_reader=lazy_reader)[0]
        else:
            y_min, y_max = y_bounds
        profile = BinnedProfile2D(data_source,
                                  x_bins, fields[0], x_min, x_max, x_log,
                                  y_bins, fields[1], y_min, y_max, y_log,
                                  lazy_reader)
        # This will add all the fields to the profile object
        if len(fields)>2:
            profile.add_fields(fields[2:], weight=weight,
                    accumulation=accumulation, fractional=fractional)

        if id is None: id = self._get_new_id()
        p = self._add_plot(PhasePlot(profile, fields, 
                                               id, cmap=cmap,
                                               figure=figure, axes=axes))
        return p

    def add_phase_sphere(self, radius, unit, fields, center = None, cmap=None,
                         weight="CellMassMsun", accumulation=False,
                         x_bins=128, x_log=True, x_bounds=None,
                         y_bins=128, y_log=True, y_bounds=None,
                         lazy_reader=True, id=None,
                         axes = None, figure = None,
                         fractional=False):
        r"""From a description of a sphere, create a 2D, binned profile.

        This function will accept the radius of a sphere, and from that it will
        generate a `Binned1DProfile`, based on the specified options.  The
        profile will be 2D, which means while it can have an arbitrary number
        of fields, those fields will all be binned based on two fields.

        All subsequent parameters beyond "unit" will be passed verbatim to
        add_profile_object.

        Parameters
        ----------
        radius : float
            The radius of the sphere to generate.
        unit : string
            The unit in which the given radius is expressed.
        fields : list of strings
            The first element of this list is the field by which we will bin
            into the y-axis, the second is the field by which we will bin onto
            the y-axis.  All subsequent fields will be binned and their
            profiles added to the underlying `BinnedProfile2D`.
        center : array_like, optional
            The center to be used for things like radius and radial velocity.
            Defaults to the center of the plot collection.
        cmap : string, optional
            An acceptable colormap.  See either yt.visualization.color_maps or
            http://www.scipy.org/Cookbook/Matplotlib/Show_colormaps .
        weight : string, default "CellMassMsun"
            The weighting field for an average.  This defaults to mass-weighted
            averaging.
        accumulation : list of booleans, optional
            If true, from the low-value to the high-value the values in all
            binned fields will be accumulated.  This is useful for instance
            when adding an unweighted CellMassMsun to a radial plot, as it will
            show mass interior to that radius.  The first value is for the
            x-axis, the second value for the y-axis.  Note that accumulation
            will only be along each row or column.
        x_bins : int, optional
            How many bins should there be in the x-axis variable?
        x_log : boolean, optional
            Should the bin edges be log-spaced?
        x_bounds : tuple of floats, optional
            If specified, the boundary values for the binning.  If unspecified,
            the min/max from the data_source will be used.  (Non-zero min/max
            in case of log-spacing.)
        y_bins : int, optional
            How many bins should there be in the y-axis variable?
        y_log : boolean, optional
            Should the bin edges be log-spaced?
        y_bounds : tuple of floats, optional
            If specified, the boundary values for the binning.  If unspecified,
            the min/max from the data_source will be used.  (Non-zero min/max
            in case of log-spacing.)
        lazy_reader : boolean, optional
            If this is false, all of the data will be read into memory before
            any processing occurs.  It defaults to true, and grids are binned
            on a one-by-one basis.  Note that parallel computation requires
            this to be true.
        id : int, optional
            If specified, this will be the "semi-unique id" of the resultant
            plot.  This should not be set.
        figure : `matplotlib.figure.Figure`, optional
            The figure onto which the axes will be placed.  Typically not used
            unless *axes* is also specified.
        axes : `matplotlib.axes.Axes`, optional
            The axes object which will be used to create the image plot.
            Typically used for things like multiplots and the like.
        fractional : boolean
            If true, the plot will be normalized to the sum of all the binned
            values.

        Returns
        -------
        plot : `yt.visualization.plot_types.PhasePlot`
            The plot that has been added to the PlotCollection.

        See Also
        --------
        yt.data_objects.profiles.BinnedProfile2D : This is the object that does the
                                   transformation of raw data into a 1D
                                   profile.

        Examples
        --------

        This will show the mass-distribution in the Density-Temperature plane.

        >>> pc.add_phase_sphere(1.0, 'kpc',
                ["Density", "Temperature", "CellMassMsun"], weight = None)
        """

        if center is None: center = self.c
        r = radius/self.pf[unit]
        data_source = self.pf.hierarchy.sphere(center, r)
        p = self.add_phase_object(data_source, fields, cmap,
                             weight, accumulation,
                             x_bins, x_log, x_bounds,
                             y_bins, y_log, y_bounds,
                             lazy_reader, id, axes=axes, figure=figure, fractional=fractional)
        p["Width"] = radius
        p["Unit"] = unit
        p["Axis"] = None
        return p

    def add_scatter_source(self, data_source, fields, id=None,
                     figure = None, axes = None, plot_options = None):
        r"""Given a data source, make a scatter plot from that data source.

        This is a very simple plot: you give it an instance of `AMRData`, two
        field names, and it will plot them on an axis

        Parameters
        ----------
        data_source : `yt.data_objects.api.AMRData`
            This will be the data source from which field values will be
            obtained.
        fields : tuple of strings
            The first of these will be the x-field, and the second the y-field.
        id : int, optional
            If specified, this will be the "semi-unique id" of the resultant
            plot.  This should not be set.
        figure : `matplotlib.figure.Figure`, optional
            The figure onto which the axes will be placed.  Typically not used
            unless *axes* is also specified.
        axes : `matplotlib.axes.Axes`, optional
            The axes object which will be used to create the image plot.
            Typically used for things like multiplots and the like.
        plot_options : dict
            These options will be given to `matplotlib.axes.Axes.scatter`
        
        Returns
        -------
        plot : `yt.visualization.plot_types.ScatterPlot`
            The plot that has been added to the PlotCollection.

        Notes
        -----
        This is a simpler way of making a phase plot, but note that because
        pixels are deposited in order, the color may be a biased sample.

        Examples
        --------

        >>> reg = pf.h.region([0.1, 0.2, 0.3], [0.0, 0.1, 0.2],
                              [0.2, 0.3, 0.4])
        >>> pc.add_scatter_plot(reg, ["Density", "Temperature"],
        >>>                     plot_options = {'color':'b'})
        """
        if id is None: id = self._get_new_id()
        if plot_options is None: plot_options = {}
        sp = ScatterPlot(data_source, fields, id,
                                   plot_options = plot_options,
                                   figure=figure, axes=axes)
        p = self._add_plot(sp)
        return p

    def add_fixed_resolution_plot(self, frb, field, use_colorbar=True,
                      figure = None, axes = None, fig_size=None):
        r"""Create a fixed resolution image from an existing buffer.

        This accepts a `FixedResolutionBuffer` and will make a plot from that
        buffer.

        Parameters
        ----------
        frb : `yt.visualization.plot_types.FixedResolutionBuffer`
            The buffer from which fields will be pulled.
        field : string
            The initial field to display.
        use_colorbar : bool, optional
            Whether we should leave room for and create a colorbar.
        figure : `matplotlib.figure.Figure`, optional
            The figure onto which the axes will be placed.  Typically not used
            unless *axes* is also specified.
        axes : `matplotlib.axes.Axes`, optional
            The axes object which will be used to create the image plot.
            Typically used for things like multiplots and the like.
        fig_size : tuple of floats
            This parameter can act as a proxy for the manual creation of a
            figure.  By specifying it, you can create plots with an arbitrarily
            large or small size.  It is in inches, defaulting to 100 dpi.

        Returns
        -------
        plot : `yt.visualization.plot_types.FixedResolutionPlot`
            The plot that has been added to the PlotCollection.

        See Also
        --------
        yt.extensions.image_writer.write_image : A faster, colorbarless way to
                                                 write out FRBs.

        Examples
        --------

        Here's a simple mechanism for getting the angular momentum of a
        collapsing cloud and generating a cutting plane aligned with the
        angular momentum vector.

        >>> pf = load("RD0005-mine/RedshiftOutput0005")
        >>> proj = pf.h.proj(0, "Density")
        >>> frb = FixedResolutionBuffer(proj, (0.2, 0.3, 0.4, 0.5), (512, 512))
        >>> p = pc.add_fixed_resolution_plot(frb, "Density")
        """
        p = self._add_plot(FixedResolutionPlot(frb, field,
                         use_colorbar=use_colorbar, axes=axes, figure=figure,
                         size=fig_size))
        p["Axis"] = "na"
        return p

    def add_ortho_ray(self, axis, coords, field, figure = None,
                      axes = None, field_parameters = None,
                      plot_options = None):
        r"""Create a ray parallel to some axis, from that a line plot, and add
        it to the current collection.

        This function will generate a `yt.data_objects.api.AMROrthoRayBase` from the given
        parameters.  This ray then gets passed to a `yt.visualization.plot_types.LineQueryPLot`, and
        the resultant plot is added to the current collection.  Various
        parameters allow control of the way the line plot is displayed, as well as
        how the ray is generated.

        Parameters
        ----------
        axis : int
            The axis along which to cast the ray.  Can be 0, 1, or 2 or x, y,
            z.
        coords : tuple of floats
            The coordinates to place the ray at.  Note that the axes are in the
            form of x_dict[axis] and y_dict[axis] for some axis.
        field : string
            The initial field to slice and display.
        figure : `matplotlib.figure.Figure`, optional
            The figure onto which the axes will be placed.  Typically not used
            unless *axes* is also specified.
        axes : `matplotlib.axes.Axes`, optional
            The axes object which will be used to create the image plot.
            Typically used for things like multiplots and the like.
        field_parameters : dict, optional
            This set of parameters will be passed to the slice upon creation,
            which can be used for passing variables to derived fields.
        plot_options : dict
            These options will be given to `matplotlib.axes.Axes.plot`

        Returns
        -------
        plot : `yt.visualization.plot_types.LineQueryPlot`
            The plot that has been added to the PlotCollection.

        See Also
        --------
        yt.data_objects.api.AMROrthoRayBase : This is the type created by this function and 
                                   passed to the plot created here.

        Examples
        --------

        This will cast a ray from (0.0, 0.5, 0.5) to (1.0, 0.5, 0.5) and plot
        it.

        >>> pf = load("RD0005-mine/RedshiftOutput0005")
        >>> pc = PlotCollection(pf, [0.5, 0.5, 0.5])
        >>> p = pc.add_ortho_ray(0, (0.5, 0.5), "Density")
        """
        axis = fix_axis(axis)
        if field_parameters is None: field_parameters = {}
        if plot_options is None: plot_options = {}
        data_source = self.pf.h.ortho_ray(axis, coords, field,
                        **field_parameters)
        p = self._add_plot(LineQueryPlot(data_source,
                [axis_names[axis], field], self._get_new_id(),
                figure=figure, axes=axes, plot_options=plot_options))
        return p

    def add_ray(self, start_point, end_point, field, figure = None,
                axes = None, field_parameters = None, plot_options = None):
        r"""Create a ray between two points, from that a line plot, and add
        it to the current collection.

        This function will generate a `yt.data_objects.api.AMRRayBase` from the given
        parameters.  This ray then gets passed to a `yt.visualization.plot_types.LineQueryPLot`, and
        the resultant plot is added to the current collection.  Various
        parameters allow control of the way the line plot is displayed, as well as
        how the ray is generated.

        Parameters
        ----------
        start_point : array_like
            The starting point of the ray.
        end_point : array_like
            The ending point of the ray.
        field : string
            The initial field to slice and display.
        figure : `matplotlib.figure.Figure`, optional
            The figure onto which the axes will be placed.  Typically not used
            unless *axes* is also specified.
        axes : `matplotlib.axes.Axes`, optional
            The axes object which will be used to create the image plot.
            Typically used for things like multiplots and the like.
        field_parameters : dict, optional
            This set of parameters will be passed to the slice upon creation,
            which can be used for passing variables to derived fields.
        plot_options : dict
            These options will be given to `matplotlib.axes.Axes.plot`

        Returns
        -------
        plot : `yt.visualization.plot_types.LineQueryPlot`
            The plot that has been added to the PlotCollection.

        See Also
        --------
        yt.data_objects.api.AMRRayBase : This is the type created by this function and 
                              passed to the plot created here.

        Examples
        --------

        This will cast a ray from (0.1, 0.2, 0.3) to (0.9, 0.7, 0.4) and plot
        it.

        >>> pf = load("RD0005-mine/RedshiftOutput0005")
        >>> pc = PlotCollection(pf, [0.5, 0.5, 0.5])
        >>> p = pc.add_ray((0.1, 0.2, 0.3), (0.9, 0.7, 0.4), "Density")
        """
        if field_parameters is None: field_parameters = {}
        if plot_options is None: plot_options = {}
        data_source = self.pf.h.ray(start_point, end_point, field,
                                    **field_parameters)
        p = self._add_plot(LineQueryPlot(data_source,
                ['t', field], self._get_new_id(),
                figure=figure, axes=axes, plot_options=plot_options))
        return p

    def _get_new_id(self):
        self.__id_counter += 1
        return self.__id_counter-1

    @rootonly
    def save_book(self, filename, author = None, title = None, keywords = None,
                  subject = None, creator = None, producer = None,
                  creation_date = None):
        r"""Save a multipage PDF of all the current plots, rather than
        individual image files.

        This function will utilize the matplotlib PDF backend to create a new
        PDF, and for every plot that the PlotCollection currently has, it will
        render a new page into that PDF.  The pages will be in the order of the
        current plots.

        Parameters
        ----------
        filename : string
            The name of the PDF file to generate.  Note that it will be
            overwritten, and '.pdf' will not be appended.
        author : string, optional
            The string to place in the metadata value of the PDF for 'author'.
        title : string, optional
            The string to place in the metadata value of the PDF for 'title'.
        keywords : string, optional
            The string to place in the metadata value of the PDF for 'keywords'.
        subject : string, optional
            The string to place in the metadata value of the PDF for 'subject'.
        creator : string, optional
            The string to place in the metadata value of the PDF for 'creator'.
        producer : string, optional
            The string to place in the metadata value of the PDF for 'producer'.
        creation_date : string, optional
            The string to place in the metadata value of the PDF for
            'creation_date'.

        Returns
        -------
        Nothing

        Examples
        --------
        This will set up a new PlotCollection, add some plots, and then save it
        as a PDF.

        >>> pc = PlotCollection(pf, [0.5, 0.5, 0.5])
        >>> pc.add_projection("Density", 'x')
        >>> pc.add_projection("Density", 'y')
        >>> pc.add_projection("Density", 'z')
        >>> pc.set_width(0.5, 'pc')
        >>> dd = pf.h.all_data()
        >>> pc.add_phase_object(dd, ["Density", "Temperature", "CellMassMsun"],
        ...                     weight = None)
        >>> pc.save_book("my_plots.pdf", author="Yours Truly",
        ...              title="Fun plots")
        """
        from matplotlib.backends.backend_pdf import PdfPages
        outfile = PdfPages(filename)
        for plot in self.plots:
            plot.save_to_pdf(outfile)
        pdf_keys = ['Title', 'Author', 'Subject', 'Keywords', 'Creator',
            'Producer', 'CreationDate']
        pdf_values = [title, author, subject, keywords, creator, producer,
            creation_date]
        metadata = outfile.infodict()
        for key, val in zip(pdf_keys, pdf_values):
            if isinstance(val, str):
                metadata[key] = val
        outfile.close()

def wrap_pylab_newplot(func):
    @wraps(func)
    def pylabify(self, *args, **kwargs):
        # Let's assume that axes and figure are not in the positional
        # arguments -- probably safe!
        new_fig = self.pylab.figure()
        try:
            new_fig.canvas.set_window_title("%s" % (self.pf))
        except AttributeError:
            pass
        if 'axes' not in kwargs: kwargs['axes'] = self.pylab.gca()
        if 'figure' not in kwargs: kwargs['figure'] = self.pylab.gcf()
        retval = func(self, *args, **kwargs)
        retval._redraw_image()
        retval._fig_num = new_fig.number
        self.pylab.draw()
        return retval
    return pylabify

def wrap_pylab_show(func):
    @wraps(func)
    def pylabify(self, *args, **kwargs):
        retval = func(self, *args, **kwargs)
        fig_num = self.pylab.gcf().number
        for p in self.plots:
            self.pylab.figure(p._fig_num)
            self.pylab.draw()
        self.pylab.figure(fig_num)
        return retval
    return pylabify

class _Interactify(type):
    # All inherited methods get wrapped if they start with add_ or set_
    # So anything inheriting this automatically gets set up; additional
    # wrappings can be done manually.  Note that this does NOT modify
    # methods that are only in the subclasses.
    def __init__(cls, name, bases, d):
        super(_Interactify, cls).__init__(name, bases, d)
        for base in bases:
            for attrname in dir(base):
                if attrname in d: continue # If overridden, don't reset
                attr = getattr(cls, attrname)
                if type(attr) == types.MethodType:
                    if attrname.startswith("add_"):
                        setattr(cls, attrname, wrap_pylab_newplot(attr))
                    elif attrname.startswith("set_"):
                        setattr(cls, attrname, wrap_pylab_show(attr))

class PlotCollectionInteractive(PlotCollection):
    __metaclass__ = _Interactify

    autoscale = wrap_pylab_show(PlotCollection.autoscale)
    switch_field = wrap_pylab_show(PlotCollection.switch_field)

    def __init__(self, *args, **kwargs):
        import pylab
        self.pylab = pylab
        super(PlotCollectionInteractive, self).__init__(*args, **kwargs)

    def redraw(self):
        r"""Redraw all affiliated plots.

        To ensure that any interactive windows are up to date, this function
        can be called to redraw all images into them.
        """
        for plot in self.plots:
            plot._redraw_image()
        self.pylab.draw()

    def clear_figures(self):
        r"""Clear all interactive figures affiliated with this collection.

        Because reusing figures between plot collections can be tricky,
        occasionally they must be manually cleared to re-obtain empty figures
        for future plotting.  This will clear all figures.
        """
        for plot in self.plots:
            self.pylab.figure(plot._fig_num)
            self.pylab.clf()

    def interactive_zoom(self):
        r"""Enter an interactive zooming session for all plots.

        Use this to enter an interactive session where zoom factors
        can be entered that are used to set the widths of the plot
        collection.  This function has no arguments, but will return
        the final width upon exit.

        Caution: Tested and works with TkAgg and MacOSX backends.  Threaded
        backends like Qt4Agg are likely to fail.

        Controls:
        Any numeric value: Zooms by this factor
        0: Exit interactive zoom session
        -1: Reset to width of 1.0 in code units
        empty: zooms by previously entered factor.

        Returns:
        width: (float) The final width of the plot collection
        """
        print 'Enter Zoom Factor, 0 to exit, -1 to reset to width=1.0'
        zfactor = 1.0
        while(True):
            new_zoom = raw_input('zoom:')
            if new_zoom is not '':
                try:
                    zfactor = float(new_zoom)
                except:
                    print 'Please enter a valid number, or 0 to exit'
                    continue
            else:
                print 'Using previous zoom value of %e' % zfactor
            if zfactor == 0.0:
                break
            elif zfactor == -1.0:
                self.set_width(1.0,'1')
            else:
                self.set_width(self.plots[0].__dict__['width']/zfactor,'1')
        print 'Returning final width of %e' % self.plots[0].width
        return self.plots[0].width

class PlotCollectionIPython(PlotCollection):
    def save(self, basename = None):
        r"""Shows all the plots hanging off this plot collection in the IPython
        web notebook.

        This function will instruct the IPython web notebook to show its
        images.

        Examples
        --------

        >>> pc.save()
        """
        from ._mpl_imports import FigureCanvasAgg
        from IPython.zmq.pylab.backend_inline import \
            send_figure
        if basename is None: basename = str(self.pf)
        for plot in self.plots:
            canvas = FigureCanvasAgg(plot._figure)
            send_figure(plot._figure)

def get_multi_plot(nx, ny, colorbar = 'vertical', bw = 4, dpi=300,
                   cbar_padding = 0.4):
    r"""Construct a multiple axes plot object, with or without a colorbar, into
    which multiple plots may be inserted.

    This will create a set of :class:`matplotlib.axes.Axes`, all lined up into
    a grid, which are then returned to the user and which can be used to plot
    multiple plots on a single figure.

    Parameters
    ----------
    nx : int
        Number of axes to create along the x-direction
    ny : int
        Number of axes to create along the y-direction
    colorbar : {'vertical', 'horizontal', None}, optional
        Should Axes objects for colorbars be allocated, and if so, should they
        correspond to the horizontal or vertical set of axes?
    bw : number
        The base height/width of an axes object inside the figure, in inches
    dpi : number
        The dots per inch fed into the Figure instantiation

    Returns
    -------
    fig : :class:`matplotlib.figure.Figure`
        The figure created inside which the axes reside
    tr : list of list of :class:`matplotlib.axes.Axes` objects
        This is a list, where the inner list is along the x-axis and the outer
        is along the y-axis
    cbars : list of :class:`matplotlib.axes.Axes` objects
        Each of these is an axes onto which a colorbar can be placed.

    Notes
    -----
    This is a simple implementation for a common use case.  Viewing the source
    can be instructure, and is encouraged to see how to generate more
    complicated or more specific sets of multiplots for your own purposes.
    """
    hf, wf = 1.0/ny, 1.0/nx
    fudge_x = fudge_y = 1.0
    if colorbar is None:
        fudge_x = fudge_y = 1.0
    elif colorbar.lower() == 'vertical':
        fudge_x = nx/(cbar_padding+nx)
        fudge_y = 1.0
    elif colorbar.lower() == 'horizontal':
        fudge_x = 1.0
        fudge_y = ny/(cbar_padding+ny)
    fig = figure.Figure((bw*nx/fudge_x, bw*ny/fudge_y), dpi=dpi)
    from _mpl_imports import FigureCanvasAgg
    fig.set_canvas(FigureCanvasAgg(fig))
    fig.subplots_adjust(wspace=0.0, hspace=0.0,
                        top=1.0, bottom=0.0,
                        left=0.0, right=1.0)
    tr = []
    print fudge_x, fudge_y
    for j in range(ny):
        tr.append([])
        for i in range(nx):
            left = i*wf*fudge_x
            bottom = fudge_y*(1.0-(j+1)*hf) + (1.0-fudge_y)
            ax = fig.add_axes([left, bottom, wf*fudge_x, hf*fudge_y])
            tr[-1].append(ax)
    cbars = []
    if colorbar is None:
        pass
    elif colorbar.lower() == 'horizontal':
        for i in range(nx):
            # left, bottom, width, height
            # Here we want 0.10 on each side of the colorbar
            # We want it to be 0.05 tall
            # And we want a buffer of 0.15
            ax = fig.add_axes([wf*(i+0.10)*fudge_x, hf*fudge_y*0.20,
                               wf*(1-0.20)*fudge_x, hf*fudge_y*0.05])
            cbars.append(ax)
    elif colorbar.lower() == 'vertical':
        for j in range(ny):
            ax = fig.add_axes([wf*(nx+0.05)*fudge_x, hf*fudge_y*(ny-(j+0.95)),
                               wf*fudge_x*0.05, hf*fudge_y*0.90])
            ax.clear()
            cbars.append(ax)
    return fig, tr, cbars

def _MPLFixImage(data_source, image_obj, field, cbar, cls):
    nx, ny = image_obj.get_size()
    def f(axes):
        x0, x1 = axes.get_xlim()
        y0, y1 = axes.get_ylim()
        frb = cls(data_source, (x0, x1, y0, y1), (nx, ny))
        image_obj.set_data(frb[field])
        mi, ma = frb[field].min(), frb[field].max()
        cbar.norm.autoscale((mi, ma))
        image_obj.set_extent([x0, x1, y0, y1])
        cbar.update_bruteforce(image_obj)
    return f

def matplotlib_widget(data_source, field, npix):
    r"""Create a widget from a data_source that uses the Matplotlib interaction
    method to pan, zoom, and so on.

    This is a simple way to take a yt data source, for instance a projection or
    a slice, and to create a matplotlib view into it that you can pan and zoom.
    It uses the matplotlib interaction engine to manage input and display.

    Parameters
    ----------
    data_source : :class:`yt.data_objects.data_containers.AMRProjBase` or :class:`yt.data_objects.data_containers.AMRSliceBase`
        This is the source to be pixelized, which can be a projection or a
        slice.  
    field : string
        The field that you want to display in the window.
    npix : int
        The number of pixels on a side you want the image to be.

    Examples
    --------

    >>> pf = load("DD0030/DD0030")
    >>> p = pf.h.proj(0, "Density")
    >>> matplotlib_widget(p, "Density", 1024)

    """
    import pylab
    import matplotlib.colors
    from .fixed_resolution import FixedResolutionBuffer, \
            ObliqueFixedResolutionBuffer
    pf = data_source.pf
    if getattr(data_source, "axis", 4) < 3:
        cls = FixedResolutionBuffer
        ax = data_source.axis
        extent = [pf.domain_left_edge[x_dict[ax]],
                  pf.domain_right_edge[x_dict[ax]],
                  pf.domain_left_edge[y_dict[ax]],
                  pf.domain_right_edge[y_dict[ax]]]
    else:
        cls = ObliqueFixedResolutionBuffer
        extent = [0.0, 1.0, 0.0, 1.0]
    take_log = pf.field_info[field].take_log
    if take_log:
        norm = matplotlib.colors.LogNorm()
    else:
        norm = matplotlib.colors.Normalize()
    ax = pylab.figure().gca()
    ax.autoscale(False)
    axi = ax.imshow(np.random.random((npix, npix)),
                    extent = extent, norm = norm,
                    origin = 'lower')
    cb = pylab.colorbar(axi, norm = norm)
    showme = _MPLFixImage(data_source, axi, field, cb, cls)
    ax.callbacks.connect("xlim_changed", showme)
    ax.callbacks.connect("ylim_changed", showme)
    ax.set_xlim(extent[0], extent[1])
    ax.set_ylim(extent[2], extent[3])
    return ax



import matplotlib
matplotlib.rc('contour', negative_linestyle='solid')

import matplotlib.image
import matplotlib.ticker
import matplotlib.axes
import matplotlib.figure
import matplotlib._image
import matplotlib.colors
import matplotlib.colorbar
import matplotlib.cm
import matplotlib.collections

from matplotlib.backends.backend_agg import \
    FigureCanvasAgg

from matplotlib.backends.backend_pdf import \
    FigureCanvasPdf

from matplotlib.backends.backend_ps import \
    FigureCanvasPS

# Now we provide some convenience functions to get information about plots.
# With Matplotlib 0.98.x, the 'transforms' branch broke backwards
# compatibility.  Despite that, the various packagers are plowing ahead with
# packaging 0.98.x with new distributions of python software.  So I guess
# we have to support it.

_compatibility_functions = ["mpl_get_bounds","mpl_notify"]

_mpl98_mpl_get_bounds = lambda bbox: bbox.bounds
_mpl9x_mpl_get_bounds = lambda bbox: bbox.get_bounds()
_mpl98_mpl_notify = lambda im,cb: cb.update_bruteforce(im)
_mpl9x_mpl_notify = lambda im,cb: cb.notify(im)

# This next function hurts, because it relies on the fact that we're
# only differentiating between 0.9[01] and 0.98. And if happens to be
# 1.0, or any version with only 3 values, this should catch it.

try:
    _mpl_version = float(matplotlib.__version__[:4])
except:
    _mpl_version = float(matplotlib.__version__[:3])

if _mpl_version < 0.98:
    _prefix = '_mpl9x'
else:
    _prefix = '_mpl98'

for fn in _compatibility_functions:
    exec("%s = %s_%s" % (fn, _prefix, fn))



"""
This is a simple mechanism for interfacing with Profile and Phase plots

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import base64
import types

from functools import wraps
import numpy as np

from .image_writer import \
    write_image, apply_colormap
from yt.utilities.lib import \
    write_png_to_string
from yt.data_objects.profiles import \
    BinnedProfile1D, \
    BinnedProfile2D
from .tick_locators import LogLocator, LinearLocator
from yt.utilities.logger import ytLogger as mylog

def invalidate_plot(f):
    @wraps(f)
    def newfunc(*args, **kwargs):
        args[0]._plot_valid = False
        args[0]._setup_plot()
        return f(*args, **kwargs)
    return newfunc

class AxisSpec(object):
    title = None
    bounds = None
    scale = None
    ticks = None

    def calculate_ticks(self):
        if self.scale == 'log':
            locator = LogLocator()
        elif self.scale == 'linear':
            locator = LinearLocator()
        else:
            raise NotImplementedError
        self.ticks = locator(*self.bounds)

class ColorbarSpec(AxisSpec):
    cmap = None
    display = True

class PlotContainer(object):
    x_spec = None
    y_spec = None
    x_values = None
    y_values = None
    plot_spec = None

    def to_mpl(self, place = None):
        import _mpl_imports as mpl
        if isinstance(place, mpl.matplotlib.figure.Figure):
            figure, place = place, None
            place = None
        else:
            figure = mpl.matplotlib.figure.Figure((10,8))
        if isinstance(place, mpl.matplotlib.axes.Axes):
            axes, place = place, None
        else:
            axes = figure.add_subplot(1,1,1)
        if self.x_spec.scale == 'log' and \
           self.y_spec.scale == 'log':
            func = axes.loglog
        elif self.x_spec.scale == 'log':
            func = axes.semilogx
        elif self.y_spec.scale == 'log':
            func = axes.semilogy
        else:
            func = axes.plot
        if self.plot_spec is None:
            kwargs = {}
        else:
            kwargs = self.plot_spec
        func(self.x_values, self.y_values, **kwargs)
        if self.x_spec.title is not None:
            axes.set_xlabel(self.x_spec.title)
        if self.y_spec.title is not None:
            axes.set_ylabel(self.y_spec.title)
        if isinstance(place, types.StringTypes):
            canvas = mpl.FigureCanvasAgg(figure)
            canvas.print_figure(place)
        return figure, axes

class ImagePlotContainer(object):
    x_spec = None
    y_spec = None
    image = None
    cbar = None

    def to_mpl(self, place = None):
        import _mpl_imports as mpl
        if isinstance(place, mpl.matplotlib.figure.Figure):
            figure, place = place, None
            place = None
        else:
            figure = mpl.matplotlib.figure.Figure((10,8))
        if isinstance(place, mpl.matplotlib.axes.Axes):
            axes, place = place, None
        else:
            axes = figure.add_subplot(1,1,1)
        # We'll go with a mesh here, even if it's inappropriate
        use_mesh = False
        xmi, xma = self.x_spec.bounds
        if self.x_spec.scale == 'log':
            x_bins = np.logspace(np.log10(xmi), np.log10(xma),
                                 self.image.shape[0]+1)
            use_mesh = True
        else:
            x_bins = np.logspace(xmi, xma, self.image.shape[0]+1)

        ymi, yma = self.y_spec.bounds
        if self.y_spec.scale == 'log':
            y_bins = np.logspace(np.log10(ymi), np.log10(yma),
                                 self.image.shape[0]+1)
            use_mesh = True
        else:
            y_bins = np.logspace(ymi, yma, self.image.shape[0]+1)

        im = self.image
        if self.cbar.scale == 'log':
            norm = mpl.matplotlib.colors.LogNorm()
        else:
            norm = mpl.matplotlib.colors.Normalize()
        if use_mesh:
            pcm = axes.pcolormesh(x_bins, y_bins, self.image, norm=norm,
                                  shading='flat', cmap = self.cbar.cmap,
                                  rasterized=True)
            if self.x_spec.scale == 'log': axes.set_xscale("log")
            if self.y_spec.scale == 'log': axes.set_yscale("log")
        else:
            axes.imshow(self.image, origin='lower', interpolation='nearest',
                        cmap = self.cbar.cmap, extent = [xmi,xma,ymi,yma],
                        norm = norm)
        if self.x_spec.title is not None:
            axes.set_xlabel(self.x_spec.title)
        if self.y_spec.title is not None:
            axes.set_ylabel(self.y_spec.title)
        if isinstance(place, types.StringTypes):
            canvas = mpl.FigureCanvasAgg(figure)
            canvas.print_figure(place)
        return figure, axes

class PhasePlotter(object):
    scale = None
    _current_field = None

    def __init__(self, data_source, field_x, field_y, field_z,
                 weight="CellMassMsun", accumulation=False,
                 x_bins=128, x_log=True, x_bounds=None,
                 y_bins=128, y_log=True, y_bounds=None,
                 lazy_reader=True, fractional=False):
        r"""From an existing object, create a 2D, binned profile.

        This function will accept an existing `AMRData` source and from that,
        it will generate a `Binned2DProfile`, based on the specified options.
        This is useful if you have extracted a region, or if you wish to bin
        some set of massages data -- or even if you wish to bin anything other
        than a sphere.  The profile will be 2D, which means while it can have
        an arbitrary number of fields, those fields will all be binned based on
        two fields.

        Parameters
        ----------
        data_source : `yt.data_objects.api.AMRData`
            This is a data source respecting the `AMRData` protocol (i.e., it
            has grids and so forth) that will be used as input to the profile
            generation.
        fields : list of strings
            The first element of this list is the field by which we will bin
            into the x-axis, the second is the field by which we will bin onto
            the y-axis.  All subsequent fields will be binned and their
            profiles added to the underlying `BinnedProfile2D`.
        cmap : string, optional
            An acceptable colormap.  See either yt.visualization.color_maps or
            http://www.scipy.org/Cookbook/Matplotlib/Show_colormaps .
        weight : string, default "CellMassMsun"
            The weighting field for an average.  This defaults to mass-weighted
            averaging.
        accumulation : list of booleans, optional
            If true, from the low-value to the high-value the values in all
            binned fields will be accumulated.  This is useful for instance
            when adding an unweighted CellMassMsun to a radial plot, as it will
            show mass interior to that radius.  The first value is for the
            x-axis, the second value for the y-axis.  Note that accumulation
            will only be along each row or column.
        x_bins : int, optional
            How many bins should there be in the x-axis variable?
        x_log : boolean, optional
            Should the bin edges be log-spaced?
        x_bounds : tuple of floats, optional
            If specified, the boundary values for the binning.  If unspecified,
            the min/max from the data_source will be used.  (Non-zero min/max
            in case of log-spacing.)
        y_bins : int, optional
            How many bins should there be in the y-axis variable?
        y_log : boolean, optional
            Should the bin edges be log-spaced?
        y_bounds : tuple of floats, optional
            If specified, the boundary values for the binning.  If unspecified,
            the min/max from the data_source will be used.  (Non-zero min/max
            in case of log-spacing.)
        lazy_reader : boolean, optional
            If this is false, all of the data will be read into memory before
            any processing occurs.  It defaults to true, and grids are binned
            on a one-by-one basis.  Note that parallel computation requires
            this to be true.
        fractional : boolean
            If true, the plot will be normalized to the sum of all the binned
            values.

        Returns
        -------
        plot : `yt.visualization.plot_types.PlotTypes.PhasePlot`
            The plot that has been added to the PlotCollection.

        See Also
        --------
        yt.data_objects.profiles.BinnedProfile2D : This is the object that does the
                                   transformation of raw data into a 1D
                                   profile.
        
        Examples
        --------
        This will show the mass-distribution in the Density-Temperature plane.
        
        >>> pf = load("RD0005-mine/RedshiftOutput0005")
        >>> reg = pf.h.region([0.1, 0.2, 0.3], [0.0, 0.1, 0.2],
        ...                   [0.2, 0.3, 0.4])
        >>> pc.add_phase_object(reg, ["Density", "Temperature", "CellMassMsun"],
        ...                     weight = None)
        """
        if x_bounds is None:
            x_min, x_max = data_source.quantities["Extrema"](
                                    field_x, non_zero = x_log,
                                    lazy_reader=lazy_reader)[0]
        else:
            x_min, x_max = x_bounds
        if y_bounds is None:
            y_min, y_max = data_source.quantities["Extrema"](
                                    field_y, non_zero = y_log,
                                    lazy_reader=lazy_reader)[0]
        else:
            y_min, y_max = y_bounds
        profile = BinnedProfile2D(data_source,
                                  x_bins, field_x, x_min, x_max, x_log,
                                  y_bins, field_y, y_min, y_max, y_log,
                                  lazy_reader)
        # This is a fallback, in case we forget.
        if field_z.startswith("CellMass") or \
           field_z.startswith("CellVolume"):
            mylog.warning("Setting weight to None")
            weight = None
        self._initial_weight = weight
        profile.add_fields(field_z, weight=weight, accumulation=accumulation, fractional=fractional)
        self._current_field = field_z
        self.profile = profile
        self.scale = {True:'log', False:'linear'}.get(
                data_source.pf.field_info[field_z].take_log, "log")
        self._setup_plot()

    def _setup_plot(self):
        xax = AxisSpec()
        xax.title = self.profile.x_bin_field
        xax.bounds = (self.profile._x_bins[0],
                      self.profile._x_bins[-1])
        xax.scale = {True: 'log', False: 'linear'}[self.profile._x_log]
        xax.calculate_ticks()

        yax = AxisSpec()
        yax.title = self.profile.y_bin_field
        yax.bounds = (self.profile._y_bins[0],
                      self.profile._y_bins[-1])
        yax.scale = {True: 'log', False: 'linear'}[self.profile._y_log]
        yax.calculate_ticks()

        cbar = ColorbarSpec()
        cbar.title = self._current_field
        if self.scale == 'log':
            nz = (self.profile[self._current_field] > 0)
            mi = self.profile[self._current_field][nz].min()
        else:
            mi = self.profile[self._current_field].min()
        ma = self.profile[self._current_field].max()
        cbar.bounds = (mi, ma)
        cbar.cmap = 'algae'
        cbar.scale = self.scale
        cbar.calculate_ticks()

        self.plot = ImagePlotContainer()
        self.plot.image = self.profile[self._current_field].transpose()
        self.plot.x_spec = xax
        self.plot.y_spec = yax
        self.plot.cbar = cbar

class PhasePlotterExtWidget(PhasePlotter):
    _ext_widget_id = None
    _widget_name = "phase_plot"

    def _setup_plot(self):
        if self._ext_widget_id is None: return
        PhasePlotter._setup_plot(self)
        # Now self.plot exists
        from yt.gui.reason.bottle_mods import PayloadHandler
        ph = PayloadHandler()
        # We set up an x axis, y axis, colorbar, and image
        xax = self._convert_axis(self.plot.x_spec)
        yax = self._convert_axis(self.plot.y_spec)
        cbar = self._convert_axis(self.plot.cbar)
        cbar['cmap_image'] = self._get_cbar_image()
        # This is a historical artifact
        raw_data = self.plot.image[::-1,:]

        if self.plot.cbar.scale == 'log':
            func = np.log10
        else:
            func = lambda a: a
        raw_data = np.repeat(raw_data, 3, axis=0)
        raw_data = np.repeat(raw_data, 3, axis=1)
        to_plot = apply_colormap(raw_data, self.plot.cbar.bounds,
                                 self.plot.cbar.cmap, func)
        if self.plot.cbar.scale == 'log':
            # Now we white-out all those regions
            #import pdb;pdb.set_trace()
            to_plot[raw_data == 0.0,:] = 255
        pngs = write_png_to_string(to_plot)
        img_data = base64.b64encode(pngs)
        payload = {'xax':xax, 'yax':yax, 'cbar':cbar,
                   'type': 'widget_payload', 'widget_id': self._ext_widget_id,
                   'image_data': img_data}
        ph.add_payload(payload)

    def _convert_ticks(self, tick_locs, bounds, func, height = 400):
        # height can be a length too; doesn't quite matter.
        mi, ma = func(bounds)
        ticks = []
        for v1,v2 in zip(tick_locs, func(tick_locs)):
            if v2 < mi or v2 > ma: continue
            p = height - height * (v2 - mi)/(ma - mi)
            ticks.append((p,v1,v2))
            #print v1, v2, mi, ma, height, p
        return ticks

    def _convert_axis(self, spec):
        func = lambda a: a
        if spec.scale == 'log': func = np.log10
        tick_info = self._convert_ticks(spec.ticks, spec.bounds, func)
        ax = {'ticks':tick_info,
              'title': spec.title}
        return ax

    def _get_cbar_image(self, height = 400, width = 40):
        # Right now there's just the single 'cmap', but that will eventually
        # change.  I think?
        vals = np.mgrid[1:0:height * 1j] * np.ones(width)[:,None]
        vals = vals.transpose()
        to_plot = apply_colormap(vals)
        pngs = write_png_to_string(to_plot)
        img_data = base64.b64encode(pngs)
        return img_data

class ProfilePlotter(object):
    scale = None
    _current_field = None

    def __init__(self, data_source, field_x, field_y, 
                 weight="CellMassMsun", accumulation=False,
                 x_bins=128, x_log=True, x_bounds=None,
                 lazy_reader=True, fractional=False):
        if x_bounds is None:
            x_min, x_max = data_source.quantities["Extrema"](
                                    field_x, non_zero = x_log,
                                    lazy_reader=lazy_reader)[0]
        else:
            x_min, x_max = x_bounds
        profile = BinnedProfile1D(data_source,
                                  x_bins, field_x, x_min, x_max, x_log,
                                  lazy_reader)
        # This is a fallback, in case we forget.
        if field_y.startswith("CellMass") or \
           field_y.startswith("CellVolume"):
            mylog.warning("Setting weight to None")
            weight = None
        profile.add_fields(field_y, weight=weight, accumulation=accumulation, fractional=fractional)
        self._current_field = field_y
        self.profile = profile
        self.scale = {True:'log', False:'linear'}.get(
                data_source.pf.field_info[field_y].take_log, "log")
        self._setup_plot()

    def _setup_plot(self):
        xax = AxisSpec()
        xax.title = self.profile.bin_field
        xax.bounds = (self.profile._bins[0],
                      self.profile._bins[-1])
        xax.scale = {True: 'log', False: 'linear'}[self.profile._x_log]
        xax.calculate_ticks()

        yax = AxisSpec()
        yax.title = self._current_field
        if self.scale == 'log':
            ny = (self.profile[self._current_field] > 0)
            mi = self.profile[self._current_field][ny].min()
        else:
            mi = self.profile[self._current_field].min()
        ma = self.profile[self._current_field].max()
        yax.bounds = (mi, ma)
        yax.scale = self.scale
        yax.calculate_ticks()

        self.plot = PlotContainer()
        self.plot.x_values = self.profile._bins
        self.plot.y_values = self.profile[self._current_field]
        self.plot.x_spec = xax
        self.plot.y_spec = yax



"""
FLASH-specific fields

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: UCSD
Homepage: http://yt-project.org/
License:
  Copyright (C) 2010-2012 Matthew Turk, John ZuHone, Anthony Scopatz.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from yt.utilities.exceptions import *
from yt.data_objects.field_info_container import \
    FieldInfoContainer, \
    NullFunc, \
    TranslationFunc, \
    FieldInfo, \
    ValidateParameter, \
    ValidateDataField, \
    ValidateProperty, \
    ValidateSpatial, \
    ValidateGridType
import yt.data_objects.universal_fields
from yt.utilities.physical_constants import \
    kboltz, mh, Na
KnownFLASHFields = FieldInfoContainer()
add_flash_field = KnownFLASHFields.add_field

FLASHFieldInfo = FieldInfoContainer.create_with_fallback(FieldInfo)
add_field = FLASHFieldInfo.add_field

# Common fields in FLASH: (Thanks to John ZuHone for this list)
#
# dens gas mass density (g/cc) --
# eint internal energy (ergs/g) --
# ener total energy (ergs/g), with 0.5*v^2 --
# gamc gamma defined as ratio of specific heats, no units
# game gamma defined as in , no units
# gpol gravitational potential from the last timestep (ergs/g)
# gpot gravitational potential from the current timestep (ergs/g)
# grac gravitational acceleration from the current timestep (cm s^-2)
# pden particle mass density (usually dark matter) (g/cc)
# pres pressure (erg/cc)
# temp temperature (K) --
# velx velocity x (cm/s) --
# vely velocity y (cm/s) --
# velz velocity z (cm/s) --

translation_dict = {"x-velocity": "velx",
                    "y-velocity": "vely",
                    "z-velocity": "velz",
                    "Density": "dens",
                    "Temperature": "temp",
                    "Pressure" : "pres", 
                    "Grav_Potential" : "gpot",
                    "particle_position_x" : "particle_posx",
                    "particle_position_y" : "particle_posy",
                    "particle_position_z" : "particle_posz",
                    "particle_velocity_x" : "particle_velx",
                    "particle_velocity_y" : "particle_vely",
                    "particle_velocity_z" : "particle_velz",
                    "particle_index" : "particle_tag",
                    "Electron_Fraction" : "elec",
                    "HI_Fraction" : "h   ",
                    "HD_Fraction" : "hd  ",
                    "HeI_Fraction": "hel ",
                    "HeII_Fraction": "hep ",
                    "HeIII_Fraction": "hepp",
                    "HM_Fraction": "hmin",
                    "HII_Fraction": "hp  ",
                    "H2I_Fraction": "htwo",
                    "H2II_Fraction": "htwp",
                    "DI_Fraction": "deut",
                    "DII_Fraction": "dplu",
                    "ParticleMass": "particle_mass",
                    "Flame_Fraction": "flam"}

def _get_density(fname):
    def _dens(field, data):
        return data[fname] * data['Density']
    return _dens

for fn1, fn2 in translation_dict.items():
    if fn1.endswith("_Fraction"):
        add_field(fn1.split("_")[0] + "_Density",
                  function=_get_density(fn1), take_log=True,
                  display_name="%s\/Density" % fn1.split("_")[0],
                  units = r"\rm{g}/\rm{cm}^{3}",
                  projected_units = r"\rm{g}/\rm{cm}^{2}",
                  )

def _get_convert(fname):
    def _conv(data):
        return data.convert(fname)
    return _conv

add_flash_field("dens", function=NullFunc, take_log=True,
                convert_function=_get_convert("dens"),
                units=r"\rm{g}/\rm{cm}^{3}",
                projected_units = r"\rm{g}/\rm{cm}^{2}"),
add_flash_field("velx", function=NullFunc, take_log=False,
                convert_function=_get_convert("velx"),
                units=r"\rm{cm}/\rm{s}")
add_flash_field("vely", function=NullFunc, take_log=False,
                convert_function=_get_convert("vely"),
                units=r"\rm{cm}/\rm{s}")
add_flash_field("velz", function=NullFunc, take_log=False,
                convert_function=_get_convert("velz"),
                units=r"\rm{cm}/\rm{s}")
add_flash_field("ener", function=NullFunc, take_log=True,
                convert_function=_get_convert("ener"),
                units=r"\rm{erg}/\rm{g}")
add_flash_field("eint", function=NullFunc, take_log=True,
                convert_function=_get_convert("eint"),
                units=r"\rm{erg}/\rm{g}")
add_flash_field("particle_posx", function=NullFunc, take_log=False,
                convert_function=_get_convert("particle_posx"),
                units=r"\rm{cm}", particle_type=True)
add_flash_field("particle_posy", function=NullFunc, take_log=False,
                convert_function=_get_convert("particle_posy"),
                units=r"\rm{cm}", particle_type=True)
add_flash_field("particle_posz", function=NullFunc, take_log=False,
                convert_function=_get_convert("particle_posz"),
                units=r"\rm{cm}", particle_type=True)
add_flash_field("particle_velx", function=NullFunc, take_log=False,
                convert_function=_get_convert("particle_velx"),
                units=r"\rm{cm}/\rm{s}", particle_type=True)
add_flash_field("particle_vely", function=NullFunc, take_log=False,
                convert_function=_get_convert("particle_vely"),
                units=r"\rm{cm}/\rm{s}", particle_type=True)
add_flash_field("particle_velz", function=NullFunc, take_log=False,
                convert_function=_get_convert("particle_velz"),
                units=r"\rm{cm}/\rm{s}", particle_type=True)
add_flash_field("particle_tag", function=NullFunc, take_log=False,
                convert_function=_get_convert("particle_tag"),
                particle_type=True)
add_flash_field("particle_mass", function=NullFunc, take_log=False,
                convert_function=_get_convert("particle_mass"),
                units=r"\rm{g}", particle_type=True)
add_flash_field("temp", function=NullFunc, take_log=True,
                convert_function=_get_convert("temp"),
                units=r"\rm{K}")
add_flash_field("tion", function=NullFunc, take_log=True,
                units=r"\rm{K}")
add_flash_field("tele", function=NullFunc, take_log=True,
                convert_function=_get_convert("tele"),
                units = r"\rm{K}")
add_flash_field("trad", function=NullFunc, take_log=True,
                units = r"\rm{K}")
add_flash_field("pres", function=NullFunc, take_log=True,
                convert_function=_get_convert("pres"),
                units=r"\rm{erg}/\rm{cm}^{3}")
add_flash_field("pion", function=NullFunc, take_log=True,
                display_name="Ion Pressure",
                units=r"\rm{erg}/\rm{cm}^3")
add_flash_field("pele", function=NullFunc, take_log=True,
                display_name="Electron Pressure, P_e",
                units=r"\rm{erg}/\rm{cm}^3")
add_flash_field("prad", function=NullFunc, take_log=True,
                display_name="Radiation Pressure",
                units = r"\rm{erg}/\rm{cm}^3")
add_flash_field("eion", function=NullFunc, take_log=True,
                display_name="Ion Internal Energy",
                units=r"\rm{erg}")
add_flash_field("eele", function=NullFunc, take_log=True,
                display_name="Electron Internal Energy",
                units=r"\rm{erg}")
add_flash_field("erad", function=NullFunc, take_log=True,
                display_name="Radiation Internal Energy",
                units=r"\rm{erg}")
add_flash_field("pden", function=NullFunc, take_log=True,
                convert_function=_get_convert("pden"),
                units=r"\rm{g}/\rm{cm}^{3}")
add_flash_field("depo", function=NullFunc, take_log=True,
                units = r"\rm{ergs}/\rm{g}")
add_flash_field("ye", function=NullFunc, take_log=True,)
add_flash_field("magx", function=NullFunc, take_log=False,
                convert_function=_get_convert("magx"),
                units = r"\mathrm{Gau\ss}")
add_flash_field("magy", function=NullFunc, take_log=False,
                convert_function=_get_convert("magy"),
                units = r"\mathrm{Gau\ss}")
add_flash_field("magz", function=NullFunc, take_log=False,
                convert_function=_get_convert("magz"),
                units = r"\mathrm{Gau\ss}")
add_flash_field("magp", function=NullFunc, take_log=True,
                convert_function=_get_convert("magp"),
                units = r"\rm{erg}/\rm{cm}^{3}")
add_flash_field("divb", function=NullFunc, take_log=False,
                convert_function=_get_convert("divb"),
                units = r"\mathrm{Gau\ss}\/\rm{cm}")
add_flash_field("game", function=NullFunc, take_log=False,
                convert_function=_get_convert("game"),
                units=r"\rm{ratio\/of\/specific\/heats}")
add_flash_field("gamc", function=NullFunc, take_log=False,
                convert_function=_get_convert("gamc"),
                units=r"\rm{ratio\/of\/specific\/heats}")
add_flash_field("gpot", function=NullFunc, take_log=False,
                convert_function=_get_convert("gpot"),
                units=r"\rm{ergs}/\rm{g}")
add_flash_field("gpol", function=NullFunc, take_log=False,
                convert_function=_get_convert("gpol"),
                units = r"\rm{ergs}/\rm{g}")
add_flash_field("flam", function=NullFunc, take_log=False,
                convert_function=_get_convert("flam"))
add_flash_field("absr", function=NullFunc, take_log=False,
                display_name="Absorption Coefficient")
add_flash_field("emis", function=NullFunc, take_log=False,
                display_name="Emissivity")
add_flash_field("cond", function=NullFunc, take_log=False,
                display_name="Conductivity")
add_flash_field("dfcf", function=NullFunc, take_log=False,
                display_name="Diffusion Equation Scalar")
add_flash_field("fllm", function=NullFunc, take_log=False,
                display_name="Flux Limit")
add_flash_field("pipe", function=NullFunc, take_log=False,
                display_name="P_i/P_e")
add_flash_field("tite", function=NullFunc, take_log=False,
                display_name="T_i/T_e")
add_flash_field("dbgs", function=NullFunc, take_log=False,
                display_name="Debug for Shocks")
add_flash_field("cham", function=NullFunc, take_log=False,
                display_name="Chamber Material Fraction")
add_flash_field("targ", function=NullFunc, take_log=False,
                display_name="Target Material Fraction")
add_flash_field("sumy", function=NullFunc, take_log=False)
add_flash_field("mgdc", function=NullFunc, take_log=False,
                display_name="Emission Minus Absorption Diffusion Terms")

for i in range(1, 1000):
    add_flash_field("r{0:03}".format(i), function=NullFunc, take_log=False,
        display_name="Energy Group {0}".format(i))


for f,v in translation_dict.items():
    if v not in KnownFLASHFields:
        pfield = v.startswith("particle")
        add_flash_field(v, function=NullFunc, take_log=False,
                  validators = [ValidateDataField(v)],
                  particle_type = pfield)
    if f.endswith("_Fraction") :
        dname = "%s\/Fraction" % f.split("_")[0]
    else :
        dname = f                    
    ff = KnownFLASHFields[v]
    pfield = f.startswith("particle")
    add_field(f, TranslationFunc(v),
              take_log=KnownFLASHFields[v].take_log,
              units = ff._units, display_name=dname,
              projected_units = ff._projected_units,
              particle_type = pfield)

def _convertParticleMassMsun(data):
    return 1.0/1.989e33
def _ParticleMassMsun(field, data):
    return data["ParticleMass"]
add_field("ParticleMassMsun",
          function=_ParticleMassMsun, validators=[ValidateSpatial(0)],
          particle_type=True, convert_function=_convertParticleMassMsun,
          particle_convert_function=_ParticleMassMsun)

def _ThermalEnergy(fields, data) :
    try:
        return data["eint"]
    except:
        pass
    try:
        return data["Pressure"] / (data.pf["Gamma"] - 1.0) / data["Density"]
    except:
        pass
    if data.has_field_parameter("mu") :
        mu = data.get_field_parameter("mu")
    else:
        mu = 0.6
    return kboltz*data["Density"]*data["Temperature"]/(mu*mh) / (data.pf["Gamma"] - 1.0)
    
add_field("ThermalEnergy", function=_ThermalEnergy,
          units=r"\rm{ergs}/\rm{g}")

def _TotalEnergy(fields, data) :
    try:
        etot = data["ener"]
    except:
        etot = data["ThermalEnergy"] + 0.5 * (
            data["x-velocity"]**2.0 +
            data["y-velocity"]**2.0 +
            data["z-velocity"]**2.0)
    try:
        etot += data['magp']/data["Density"]
    except:
        pass
    return etot

add_field("TotalEnergy", function=_TotalEnergy,
          units=r"\rm{ergs}/\rm{g}")

def _GasEnergy(fields, data) :
    return data["ThermalEnergy"]

add_field("GasEnergy", function=_GasEnergy, 
          units=r"\rm{ergs}/\rm{g}")

# See http://flash.uchicago.edu/pipermail/flash-users/2012-October/001180.html
# along with the attachment to that e-mail for details
def GetMagRescalingFactor(pf):
    if pf['unitsystem'].lower() == "cgs":
         factor = 1
    elif pf['unitsystem'].lower() == "si":
         factor = np.sqrt(4*np.pi/1e7)
    elif pf['unitsystem'].lower() == "none":
         factor = np.sqrt(4*np.pi)
    else:
        raise RuntimeError("Runtime parameter unitsystem with "
                           "value %s is unrecognized" % pf['unitsystem'])
    return factor

def _Bx(fields, data):
    factor = GetMagRescalingFactor(data.pf)
    return data['magx']*factor
add_field("Bx", function=_Bx, take_log=False,
          units=r"\rm{Gauss}", display_name=r"B_x")

def _By(fields, data):
    factor = GetMagRescalingFactor(data.pf)
    return data['magy']*factor
add_field("By", function=_By, take_log=False,
          units=r"\rm{Gauss}", display_name=r"B_y")

def _Bz(fields, data):
    factor = GetMagRescalingFactor(data.pf)
    return data['magz']*factor
add_field("Bz", function=_Bz, take_log=False,
          units=r"\rm{Gauss}", display_name=r"B_z")

def _DivB(fields, data):
    factor = GetMagRescalingFactor(data.pf)
    return data['divb']*factor
add_field("DivB", function=_DivB, take_log=False,
          units=r"\rm{Gauss}\/\rm{cm}^{-1}")



## Derived FLASH Fields
def _nele(field, data):
    return data['dens'] * data['ye'] * Na
add_field('nele', function=_nele, take_log=True, units=r"\rm{cm}^{-3}")
add_field('edens', function=_nele, take_log=True, units=r"\rm{cm}^{-3}")

def _nion(field, data):
    return data['dens'] * data['sumy'] * Na
add_field('nion', function=_nion, take_log=True, units=r"\rm{cm}^{-3}")

def _abar(field, data):
    return 1.0 / data['sumy']
add_field('abar', function=_abar, take_log=False)



"""
Fixed interpolator includes

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

cimport numpy as np


cdef extern from "FixedInterpolator.h":
    np.float64_t fast_interpolate(int ds[3], int ci[3], np.float64_t dp[3],
                                  np.float64_t *data) nogil
    np.float64_t offset_interpolate(int ds[3], np.float64_t dp[3],
                                    np.float64_t *data) nogil
    np.float64_t trilinear_interpolate(int ds[3], int ci[3], np.float64_t dp[3],
                                       np.float64_t *data) nogil
    void eval_gradient(int ds[3], np.float64_t dp[3], np.float64_t *data,
                       np.float64_t grad[3]) nogil
    void offset_fill(int *ds, np.float64_t *data, np.float64_t *gridval) nogil
    void vertex_interp(np.float64_t v1, np.float64_t v2, np.float64_t isovalue,
                       np.float64_t vl[3], np.float64_t dds[3],
                       np.float64_t x, np.float64_t y, np.float64_t z,
                       int vind1, int vind2) nogil




"""
A mechanism for listing available analysis modules.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2007-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import sys

def get_available_modules():
    modpath = os.path.abspath(os.path.dirname(__file__))
    available_modules = []
    for d in [os.path.join(modpath, f) for f in os.listdir(modpath)]:
        if os.path.isdir(d) and os.path.isfile(os.path.join(d, "api.py")):
            available_modules.append(os.path.basename(d))
    return available_modules

class AnalysisModuleLoader(object):

    @property
    def available_modules(self):
        return get_available_modules()

    def __getattr__(self, attr):
        try:
            name = "yt.analysis_modules.%s.api" % (attr)
            nm = __import__(name, level=-1)
            setattr(self, attr, sys.modules[name])
        except ImportError:
            raise AttributeError(attr)
        return getattr(self, attr)

    def __dir__(self):
        # This is a badly behaving object.  I was unable to get this line:
        #return super(AnalysisModuleLoader, self).__dir__() + self.available_modules
        # to work, so we simply return only the methods we know about.
        return ["available_modules"] + self.available_modules

amods = AnalysisModuleLoader()



"""
ART-specific IO

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2007-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import struct

import os
import os.path

from yt.utilities.io_handler import \
    BaseIOHandler

from yt.utilities.io_handler import \
    BaseIOHandler
import yt.utilities.lib as au

from yt.frontends.art.definitions import *

class IOHandlerART(BaseIOHandler):
    _data_style = "art"

    def __init__(self, filename, nhydro_vars, level_info, level_offsets,
                 *args, **kwargs):
        BaseIOHandler.__init__(self, *args, **kwargs)
        self.filename = filename
        self.nhydro_vars = nhydro_vars
        self.level_info = level_info
        self.level_offsets = level_offsets
        self.level_data = {}

    def preload_level(self, level,field=None):
        """ Reads in the full ART tree. From the ART source:
            iOctLv :    >0   - level of an oct
            iOctPr :         - parent of an oct
            iOctCh :    >0   - pointer to an oct of children
                        0   - there are no children; the cell is a leaf
            iOctNb :    >0   - pointers to neighbouring cells 
            iOctPs :         - coordinates of Oct centers
            
            iOctLL1:         - doubly linked list of octs
            iOctLL2:         - doubly linked list of octs
            
            tl - current  time moment for level L
            tlold - previous time moment for level L
            dtl - dtime0/2**iTimeBin
            dtlold -  previous time step for level L
            iSO - sweep order
            
            hvar(1,*) - gas density 
            hvar(2,*) - gas energy 
            hvar(3,*) - x-momentum 
            hvar(4,*) - y-momentum
            hvar(5,*) - z-momentum
            hvar(6,*) - pressure
            hvar(7,*) - Gamma
            hvar(8,*) - internal energy 

            var (1,*) - total density 
            var (2,*) - potential (new)
            var (3,*) - potential (old)
            
            
            
        """
        
        if level in self.level_data: return
        if level == 0:
            self.preload_root_level()
            return
        f = open(self.filename, 'rb')
        f.seek(self.level_offsets[level])
        ncells = 8*self.level_info[level]
        nvals = ncells * (self.nhydro_vars + 6) # 2 vars, 2 pads
        arr = np.fromfile(f, dtype='>f', count=nvals)
        arr = arr.reshape((self.nhydro_vars+6, ncells), order="F")
        assert np.all(arr[0,:]==arr[-1,:]) #pads must be equal
        arr = arr[3:-1,:] #skip beginning pad, idc, iOctCh, + ending pad
        if field==None:
            self.level_data[level] = arr.astype('float32')
        else:
            self.level_data[level] = arr.astype('float32')
        del arr

    def preload_root_level(self):
        f = open(self.filename, 'rb')
        f.seek(self.level_offsets[0] + 4) # Ditch the header
        ncells = self.level_info[0]
        nhvals = ncells * (self.nhydro_vars) # 0 vars, 0 pads
        hvar = np.fromfile(f, dtype='>f', count=nhvals).astype("float32")
        hvar = hvar.reshape((self.nhydro_vars, ncells), order="F")
        np.fromfile(f,dtype='>i',count=2) #throw away the pads
        nvars = ncells * (2) # 0 vars, 0 pads
        var = np.fromfile(f, dtype='>f', count=nvars).astype("float32")
        var = var.reshape((2, ncells), order="F")
        arr = np.concatenate((hvar,var))
        self.level_data[0] = arr

    def clear_level(self, level):
        self.level_data.pop(level, None)

    def _read_particle_field(self, grid, field):
        dat = getattr(grid,field,None)
        if dat is not None: 
            return dat
        starfield = field.replace('star','particle')
        dat = getattr(grid,starfield,None)
        if dat is not None:
            psi = grid.pf.particle_star_index
            idx = grid.particle_type==psi
            return dat[idx]
        raise KeyError
        
    def _read_data(self, grid, field):
        if field in particle_fields:
            return self._read_particle_field(grid, field)
        pf = grid.pf
        field_id = grid.pf.h.field_list.index(field)
        if grid.Level == 0: # We only have one root grid
            self.preload_level(0)
            tr = self.level_data[0][field_id,:].reshape(
                    pf.domain_dimensions, order="F").copy()
            return tr.swapaxes(0, 2).astype("float64")
        tr = np.zeros(grid.ActiveDimensions, dtype='float32')
        grids = [grid]
        l_delta = 0
        filled = np.zeros(grid.ActiveDimensions, dtype='uint8')
        to_fill = grid.ActiveDimensions.prod()
        while to_fill > 0 and len(grids) > 0:
            next_grids = []
            for g in grids:
                self.preload_level(g.Level,field=field_id)
                #print "Filling %s from %s (%s)" % (grid, g, g.Level)
                to_fill -= au.read_art_grid(field_id, 
                        grid.get_global_startindex(), grid.ActiveDimensions,
                        tr, filled, self.level_data[g.Level],
                        g.Level, 2**l_delta, g.locations)
                next_grids += g.Parent
            grids = next_grids
            l_delta += 1
        return tr.astype("float64")

def _count_art_octs(f, offset, 
                   MinLev, MaxLevelNow):
    level_oct_offsets= [0,]
    level_child_offsets= [0,]
    f.seek(offset)
    nchild,ntot=8,0
    Level = np.zeros(MaxLevelNow+1 - MinLev, dtype='int64')
    iNOLL = np.zeros(MaxLevelNow+1 - MinLev, dtype='int64')
    iHOLL = np.zeros(MaxLevelNow+1 - MinLev, dtype='int64')
    for Lev in xrange(MinLev + 1, MaxLevelNow+1):
        level_oct_offsets.append(f.tell())

        #Get the info for this level, skip the rest
        #print "Reading oct tree data for level", Lev
        #print 'offset:',f.tell()
        Level[Lev], iNOLL[Lev], iHOLL[Lev] = struct.unpack(
           '>iii', _read_record(f))
        #print 'Level %i : '%Lev, iNOLL
        #print 'offset after level record:',f.tell()
        iOct = iHOLL[Lev] - 1
        nLevel = iNOLL[Lev]
        nLevCells = nLevel * nchild
        ntot = ntot + nLevel

        #Skip all the oct hierarchy data
        ns = _read_record_size(f)
        size = struct.calcsize('>i') + ns + struct.calcsize('>i')
        f.seek(f.tell()+size * nLevel)

        level_child_offsets.append(f.tell())
        #Skip the child vars data
        ns = _read_record_size(f)
        size = struct.calcsize('>i') + ns + struct.calcsize('>i')
        f.seek(f.tell()+size * nLevel*nchild)

        #find nhydrovars
        nhydrovars = 8+2
    f.seek(offset)
    return nhydrovars, iNOLL, level_oct_offsets, level_child_offsets

def _read_art_level_info(f, level_oct_offsets,level,coarse_grid=128):
    pos = f.tell()
    f.seek(level_oct_offsets[level])
    #Get the info for this level, skip the rest
    junk, nLevel, iOct = struct.unpack(
       '>iii', _read_record(f))
    
    #fortran indices start at 1
    
    #Skip all the oct hierarchy data
    le     = np.zeros((nLevel,3),dtype='int64')
    fl     = np.ones((nLevel,6),dtype='int64')
    iocts  = np.zeros(nLevel+1,dtype='int64')
    idxa,idxb = 0,0
    chunk = long(1e6) #this is ~111MB for 15 dimensional 64 bit arrays
    left = nLevel
    while left > 0 :
        this_chunk = min(chunk,left)
        idxb=idxa+this_chunk
        data = np.fromfile(f,dtype='>i',count=this_chunk*15)
        data=data.reshape(this_chunk,15)
        left-=this_chunk
        le[idxa:idxb,:] = data[:,1:4]
        fl[idxa:idxb,1] = np.arange(idxa,idxb)
        #pad byte is last, LL2, then ioct right before it
        iocts[idxa:idxb] = data[:,-3] 
        idxa=idxa+this_chunk
    del data
    
    #ioct always represents the index of the next variable
    #not the current, so shift forward one index
    #the last index isn't used
    ioctso = iocts.copy()
    iocts[1:]=iocts[:-1] #shift
    iocts = iocts[:nLevel] #chop off the last index
    iocts[0]=iOct #starting value

    #now correct iocts for fortran indices start @ 1
    iocts = iocts-1

    assert np.unique(iocts).shape[0] == nLevel
    
    #ioct tries to access arrays much larger than le & fl
    #just make sure they appear in the right order, skipping
    #the empty space in between
    idx = np.argsort(iocts)
    
    #now rearrange le & fl in order of the ioct
    le = le[idx]
    fl = fl[idx]


    #left edges are expressed as if they were on 
    #level 15, so no matter what level max(le)=2**15 
    #correct to the yt convention
    #le = le/2**(root_level-1-level)-1

    #try to find the root_level first
    root_level=np.floor(np.log2(le.max()*1.0/coarse_grid))
    root_level = root_level.astype('int64')

    #try without the -1
    le = le/2**(root_level+1-level)-1

    #now read the hvars and vars arrays
    #we are looking for iOctCh
    #we record if iOctCh is >0, in which it is subdivided
    iOctCh  = np.zeros((nLevel+1,8),dtype='bool')
    
    
    
    f.seek(pos)
    return le,fl,nLevel,root_level


def read_particles(file,Nrow):
    words = 6 # words (reals) per particle: x,y,z,vx,vy,vz
    real_size = 4 # for file_particle_data; not always true?
    np_per_page = Nrow**2 # defined in ART a_setup.h
    num_pages = os.path.getsize(file)/(real_size*words*np_per_page)

    f = np.fromfile(file, dtype='>f4').astype('float32') # direct access
    pages = np.vsplit(np.reshape(f, (num_pages, words, np_per_page)), num_pages)
    data = np.squeeze(np.dstack(pages)).T # x,y,z,vx,vy,vz
    return data[:,0:3],data[:,3:]

def read_stars(file):
    fh = open(file,'rb')
    tdum,adum   = _read_frecord(fh,'>d')
    nstars      = _read_frecord(fh,'>i')
    ws_old, ws_oldi = _read_frecord(fh,'>d')
    mass    = _read_frecord(fh,'>f') 
    imass   = _read_frecord(fh,'>f') 
    tbirth  = _read_frecord(fh,'>f') 
    if fh.tell() < os.path.getsize(file):
        metallicity1 = _read_frecord(fh,'>f') 
    if fh.tell() < os.path.getsize(file):
        metallicity2 = _read_frecord(fh,'>f')     
    assert fh.tell() == os.path.getsize(file)
    return  nstars, mass, imass, tbirth, metallicity1, metallicity2,\
            ws_old,ws_oldi,tdum,adum

def _read_child_mask_level(f, level_child_offsets,level,nLevel,nhydro_vars):
    f.seek(level_child_offsets[level])
    nvals = nLevel * (nhydro_vars + 6) # 2 vars, 2 pads
    ioctch = np.zeros(nLevel,dtype='uint8')
    idc = np.zeros(nLevel,dtype='int32')
    
    chunk = long(1e6)
    left = nLevel
    width = nhydro_vars+6
    a,b=0,0
    while left > 0:
        chunk = min(chunk,left)
        b += chunk
        arr = np.fromfile(f, dtype='>i', count=chunk*width)
        arr = arr.reshape((width, chunk), order="F")
        assert np.all(arr[0,:]==arr[-1,:]) #pads must be equal
        idc[a:b]    = arr[1,:]-1 #fix fortran indexing
        ioctch[a:b] = arr[2,:]==0 #if it is above zero, then refined available
        #zero in the mask means there is refinement available
        a=b
        left -= chunk
    assert left==0
    return idc,ioctch
    
nchem=8+2
dtyp = np.dtype(">i4,>i8,>i8"+",>%sf4"%(nchem)+ \
                ",>%sf4"%(2)+",>i4")
def _read_art_child(f, level_child_offsets,level,nLevel,field):
    pos=f.tell()
    f.seek(level_child_offsets[level])
    arr = np.fromfile(f, dtype='>f', count=nLevel * 8)
    arr = arr.reshape((nLevel,16), order="F")
    arr = arr[3:-1,:].astype("float64")
    f.seek(pos)
    return arr[field,:]

def _skip_record(f):
    s = struct.unpack('>i', f.read(struct.calcsize('>i')))
    f.seek(s[0], 1)
    s = struct.unpack('>i', f.read(struct.calcsize('>i')))

def _read_frecord(f,fmt):
    s1 = struct.unpack('>i', f.read(struct.calcsize('>i')))[0]
    count = s1/np.dtype(fmt).itemsize
    ss = np.fromfile(f,fmt,count=count)
    s2 = struct.unpack('>i', f.read(struct.calcsize('>i')))[0]
    assert s1==s2
    return ss


def _read_record(f,fmt=None):
    s = struct.unpack('>i', f.read(struct.calcsize('>i')))[0]
    ss = f.read(s)
    s = struct.unpack('>i', f.read(struct.calcsize('>i')))
    if fmt is not None:
        return struct.unpack(ss,fmt)
    return ss

def _read_record_size(f):
    pos = f.tell()
    s = struct.unpack('>i', f.read(struct.calcsize('>i')))
    f.seek(pos)
    return s[0]

def _read_struct(f,structure,verbose=False):
    vals = {}
    for format,name in structure:
        size = struct.calcsize(format)
        (val,) = struct.unpack(format,f.read(size))
        vals[name] = val
        if verbose: print "%s:\t%s\t (%d B)" %(name,val,f.tell())
    return vals



#All of these functions are to convert from hydro time var to 
#proper time
sqrt = np.sqrt
sign = np.sign

def find_root(f,a,b,tol=1e-6):
    c = (a+b)/2.0
    last = -np.inf
    assert(sign(f(a)) != sign(f(b)))  
    while np.abs(f(c)-last) > tol:
        last=f(c)
        if sign(last)==sign(f(b)):
            b=c
        else:
            a=c
        c = (a+b)/2.0
    return c

def quad(fintegrand,xmin,xmax,n=1e4):
    spacings = np.logspace(np.log10(xmin),np.log10(xmax),n)
    integrand_arr = fintegrand(spacings)
    val = np.trapz(integrand_arr,dx=np.diff(spacings))
    return val

def a2b(at,Om0=0.27,Oml0=0.73,h=0.700):
    def f_a2b(x):
        val = 0.5*sqrt(Om0) / x**3.0
        val /= sqrt(Om0/x**3.0 +Oml0 +(1.0 - Om0-Oml0)/x**2.0)
        return val
    #val, err = si.quad(f_a2b,1,at)
    val = quad(f_a2b,1,at)
    return val

def b2a(bt,**kwargs):
    #converts code time into expansion factor 
    #if Om0 ==1and OmL == 0 then b2a is (1 / (1-td))**2
    #if bt < -190.0 or bt > -.10:  raise 'bt outside of range'
    f_b2a = lambda at: a2b(at,**kwargs)-bt
    return find_root(f_b2a,1e-4,1.1)
    #return so.brenth(f_b2a,1e-4,1.1)
    #return brent.brent(f_b2a)

def a2t(at,Om0=0.27,Oml0=0.73,h=0.700):
    integrand = lambda x : 1./(x*sqrt(Oml0+Om0*x**-3.0))
    #current_time,err = si.quad(integrand,0.0,at,epsabs=1e-6,epsrel=1e-6)
    current_time = quad(integrand,1e-4,at)
    #spacings = np.logspace(-5,np.log10(at),1e5)
    #integrand_arr = integrand(spacings)
    #current_time = np.trapz(integrand_arr,dx=np.diff(spacings))
    current_time *= 9.779/h
    return current_time

def b2t(tb,n = 1e2,logger=None,**kwargs):
    tb = np.array(tb)
    if type(tb) == type(1.1): 
        return a2t(b2a(tb))
    if tb.shape == (): 
        return a2t(b2a(tb))
    if len(tb) < n: n= len(tb)
    age_min = a2t(b2a(tb.max(),**kwargs),**kwargs)
    age_max = a2t(b2a(tb.min(),**kwargs),**kwargs)
    tbs  = -1.*np.logspace(np.log10(-tb.min()),
                          np.log10(-tb.max()),n)
    ages = []
    for i,tbi in enumerate(tbs):
        ages += a2t(b2a(tbi)),
        if logger: logger(i)
    ages = np.array(ages)
    fb2t = np.interp(tb,tbs,ages)
    #fb2t = interp1d(tbs,ages)
    return fb2t

def spread_ages(ages,logger=None,spread=1.0e7*365*24*3600):
    #stars are formed in lumps; spread out the ages linearly
    da= np.diff(ages)
    assert np.all(da<=0)
    #ages should always be decreasing, and ordered so
    agesd = np.zeros(ages.shape)
    idx, = np.where(da<0)
    idx+=1 #mark the right edges
    #spread this age evenly out to the next age
    lidx=0
    lage=0
    for i in idx:
        n = i-lidx #n stars affected
        rage = ages[i]
        lage = max(rage-spread,0.0)
        agesd[lidx:i]=np.linspace(lage,rage,n)
        lidx=i
        #lage=rage
        if logger: logger(i)
    #we didn't get the last iter
    i=ages.shape[0]-1
    n = i-lidx #n stars affected
    rage = ages[i]
    lage = max(rage-spread,0.0)
    agesd[lidx:i]=np.linspace(lage,rage,n)
    return agesd



from yt.mods import *
import matplotlib
import pylab
from output_tests import SingleOutputTest, YTStaticOutputTest, create_test
import hashlib
import numpy as np

# Tests to make sure that grid quantities are identical that should
# be identical for the AND operator.
class TestBooleanANDGridQuantity(YTStaticOutputTest):
    def run(self):
        domain = self.pf.domain_right_edge - self.pf.domain_left_edge
        four = 0.4 * domain + self.pf.domain_left_edge
        five = 0.5 * domain + self.pf.domain_left_edge
        six = 0.6 * domain + self.pf.domain_left_edge
        re1 = self.pf.h.region_strict(five, four, six)
        re2 = self.pf.h.region_strict(five, five, six)
        re = self.pf.h.boolean([re1, "AND", re2])
        # re should look like re2.
        x2 = re2['x']
        x = re['x']
        x2 = x2[x2.argsort()]
        x = x[x.argsort()]
        self.result = (x2, x)
    
    def compare(self, old_result):
        self.compare_array_delta(self.result[0], self.result[1], 1e-10)
    
    def plot(self):
        return []

# OR
class TestBooleanORGridQuantity(YTStaticOutputTest):
    def run(self):
        domain = self.pf.domain_right_edge - self.pf.domain_left_edge
        four = 0.4 * domain + self.pf.domain_left_edge
        five = 0.5 * domain + self.pf.domain_left_edge
        six = 0.6 * domain + self.pf.domain_left_edge
        re1 = self.pf.h.region_strict(five, four, six)
        re2 = self.pf.h.region_strict(five, five, six)
        re = self.pf.h.boolean([re1, "OR", re2])
        # re should look like re1
        x1 = re1['x']
        x = re['x']
        x1 = x1[x1.argsort()]
        x = x[x.argsort()]
        self.result = (x1, x)
    
    def compare(self, old_result):
        self.compare_array_delta(self.result[0], self.result[1], 1e-10)
    
    def plot(self):
        return []

# NOT
class TestBooleanNOTGridQuantity(YTStaticOutputTest):
    def run(self):
        domain = self.pf.domain_right_edge - self.pf.domain_left_edge
        four = 0.4 * domain + self.pf.domain_left_edge
        five = 0.5 * domain + self.pf.domain_left_edge
        six = 0.6 * domain + self.pf.domain_left_edge
        re1 = self.pf.h.region_strict(five, four, six)
        re2 = self.pf.h.region_strict(five, five, six)
        # Bottom base
        re3 = self.pf.h.region_strict(five, four, [six[0], six[1], five[2]])
        # Side
        re4 = self.pf.h.region_strict(five, [four[0], four[1], five[2]],
            [five[0], six[1], six[2]])
        # Last small cube
        re5 = self.pf.h.region_strict(five, [five[0], four[0], four[2]],
            [six[0], five[1], six[2]])
        # re1 NOT re2 should look like re3 OR re4 OR re5
        re = self.pf.h.boolean([re1, "NOT", re2])
        reo = self.pf.h.boolean([re3, "OR", re4, "OR", re5])
        x = re['x']
        xo = reo['x']
        x = x[x.argsort()]
        xo = xo[xo.argsort()]
        self.result = (x, xo)
    
    def compare(self, old_result):
        self.compare_array_delta(self.result[0], self.result[1], 1e-10)
    
    def plot(self):
        return []

# Tests to make sure that particle quantities are identical that should
# be identical for the AND operator.
class TestBooleanANDParticleQuantity(YTStaticOutputTest):
    def run(self):
        domain = self.pf.domain_right_edge - self.pf.domain_left_edge
        four = 0.4 * domain + self.pf.domain_left_edge
        five = 0.5 * domain + self.pf.domain_left_edge
        six = 0.6 * domain + self.pf.domain_left_edge
        re1 = self.pf.h.region_strict(five, four, six)
        re2 = self.pf.h.region_strict(five, five, six)
        re = self.pf.h.boolean([re1, "AND", re2])
        # re should look like re2.
        x2 = re2['particle_position_x']
        x = re['particle_position_x']
        x2 = x2[x2.argsort()]
        x = x[x.argsort()]
        self.result = (x2, x)
    
    def compare(self, old_result):
        self.compare_array_delta(self.result[0], self.result[1], 1e-10)
    
    def plot(self):
        return []

# OR
class TestBooleanORParticleQuantity(YTStaticOutputTest):
    def run(self):
        domain = self.pf.domain_right_edge - self.pf.domain_left_edge
        four = 0.4 * domain + self.pf.domain_left_edge
        five = 0.5 * domain + self.pf.domain_left_edge
        six = 0.6 * domain + self.pf.domain_left_edge
        re1 = self.pf.h.region_strict(five, four, six)
        re2 = self.pf.h.region_strict(five, five, six)
        re = self.pf.h.boolean([re1, "OR", re2])
        # re should look like re1
        x1 = re1['particle_position_x']
        x = re['particle_position_x']
        x1 = x1[x1.argsort()]
        x = x[x.argsort()]
        self.result = (x1, x)
    
    def compare(self, old_result):
        self.compare_array_delta(self.result[0], self.result[1], 1e-10)
    
    def plot(self):
        return []

# NOT
class TestBooleanNOTParticleQuantity(YTStaticOutputTest):
    def run(self):
        domain = self.pf.domain_right_edge - self.pf.domain_left_edge
        four = 0.4 * domain + self.pf.domain_left_edge
        five = 0.5 * domain + self.pf.domain_left_edge
        six = 0.6 * domain + self.pf.domain_left_edge
        re1 = self.pf.h.region_strict(five, four, six)
        re2 = self.pf.h.region_strict(five, five, six)
        # Bottom base
        re3 = self.pf.h.region_strict(five, four, [six[0], six[1], five[2]])
        # Side
        re4 = self.pf.h.region_strict(five, [four[0], four[1], five[2]],
            [five[0], six[1], six[2]])
        # Last small cube
        re5 = self.pf.h.region_strict(five, [five[0], four[0], four[2]],
            [six[0], five[1], six[2]])
        # re1 NOT re2 should look like re3 OR re4 OR re5
        re = self.pf.h.boolean([re1, "NOT", re2])
        reo = self.pf.h.boolean([re3, "OR", re4, "OR", re5])
        x = re['particle_position_x']
        xo = reo['particle_position_x']
        x = x[x.argsort()]
        xo = xo[xo.argsort()]
        self.result = (x, xo)
    
    def compare(self, old_result):
        self.compare_array_delta(self.result[0], self.result[1], 1e-10)
    
    def plot(self):
        return []

